# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.io.json_api_object import JsonApiObject, JsonApiError
from .json_rpc import (
    JsonRpcRequest,
    JsonRpcResponse,
    JsonRpcSuccessResponse,
    JsonRpcErrorResponse,
)
from openc3.top_level import HazardousError, CriticalCmdError, DisabledError  # noqa: F401


class JsonDrbUnknownError(Exception):
    pass


# The Ruby side implements from_hash directly on Exception but Python says:
# TypeError: cannot set 'from_hash' attribute of immutable typ 'Exception'
# So we implement it on our custom Exception


class JsonDRbError(JsonApiError):
    @classmethod
    def from_hash(cls, hash):
        # Hash contains class, message, backtrace, and instance_variables
        error_class = None
        try:
            error_class = globals()[hash["class"]]
        except KeyError:
            error_class = RuntimeError
            if "message" not in hash and "class" in hash:
                hash["message"] = hash["class"]
        error = None
        if error_class is RuntimeError and "message" in hash:
            error = error_class(hash["message"])
        else:
            error = error_class()
        if "instance_variables" in hash:
            for name, value in hash["instance_variables"].items():
                # Backend is Ruby so remove '@' from instance_variables
                setattr(error, name[1:], value)
        return error


# Used to forward all method calls to the remote server object. Before using
# this class ensure the remote service has been started in the server class:
#
#   json = JsonDrb.new
#   json.start_service('127.0.0.1', 7777, self)
#
# Now the JsonDRbObject can be used to call server methods directly:
#
#   server = JsonDRbObject('http://openc3-cosmos-cmd-tlm-api:2901', 1.0)
#   server.cmd(*args)
#
class JsonDRbObject(JsonApiObject):
    USER_AGENT = "OpenC3 / v5 (ruby/openc3/lib/io/json_drb_object)"

    # @param url [String] The url of openc3-cosmos-cmd-tlm-api http://openc3-cosmos-cmd-tlm-api:2901
    # @param timeout [Float] The time to wait before disconnecting 1.0
    # @param authentication [OpenC3Authentication] The authentication object if Nonel initialize will generate
    def __init__(self, url, timeout=1.0, authentication=None):
        super().__init__(url, timeout, authentication)
        self.uri = f"{url}/openc3-api/api"

    # Forwards all method calls to the remote service.
    #
    # @param method_name [Symbol] Name of the method to call
    # @param method_params [Array] Array of parameters to pass to the method
    # @param keyword_params [Hash<Symbol, Variable>] Hash of keyword parameters
    # @return The result of the method call. If the method raises an exception
    #   the same exception is also raised. If something goes wrong with the
    #   protocol a JsonDRbError exception is raised.
    def __getattr__(self, func):
        if self.shutdown:
            raise JsonDRbError("Shutdown")

        def method(*args, **kwargs):
            with self.mutex:
                self.log = [None, None, None]
                if not self.http:
                    self.connect()
                json_rpc_request = JsonRpcRequest(0, func, *args, **kwargs)
                token = kwargs.get("token", None)
                response_body = self.make_request(json_rpc_request, token)
                if not response_body:
                    self.disconnect()
                    error = f"No response from server: {self.log[0]} ::: {self.log[1]} ::: {self.log[2]}"
                    raise JsonDRbError(error)
                else:
                    response = JsonRpcResponse.from_hash(response_body)
                    return self.handle_response(response)

        return method

    def make_request(self, request, token=None):
        if self.authentication and not token:
            token = self.authentication.token()
        if token:
            headers = {
                "User-Agent": self.USER_AGENT,
                "Content-Type": "application/json-rpc",
                "Authorization": token,
            }
        else:
            headers = {
                "User-Agent": self.USER_AGENT,
                "Content-Type": "application/json-rpc",
            }
        try:
            request_kwargs = {
                "url": self.uri,
                "data": json.dumps(request.to_hash()),
                "headers": headers,
            }
            self.log[0] = f"Request: {request_kwargs}"
            # print(self.log[0])
            resp = self.http.post(**request_kwargs)
            self.log[1] = f"Response: {resp.status_code} {resp.headers} {resp.text}"
            # print(self.log[1])
            self.response_data = resp.json()
            return resp.json()
        except Exception as e:  # Typically JSONDecodeError when error in resp.json()
            self.log[2] = f"Exception: {repr(e)}"
            return None

    def handle_response(self, response: JsonRpcSuccessResponse | JsonRpcErrorResponse):
        # The code below will always either raise or return breaking out of the loop
        if isinstance(response, JsonRpcErrorResponse):
            if response.error.data:
                error = JsonDRbError.from_hash(response.error.data)
                raise error
            else:
                raise RuntimeError(f"JsonDRb Error ({response})")
        else:
            return response.result



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import socket
import errno
from openc3.streams.tcpip_socket_stream import TcpipSocketStream
from openc3.config.config_parser import ConfigParser


# Data {Stream} which reads and writes to TCPIP sockets. This class creates
# the actual sockets based on the constructor parameters. The rest of the
# interface is implemented by the super class {TcpipSocketStream}.
class TcpipClientStream(TcpipSocketStream):
    # self.param hostname [String] The host to connect to
    # self.param write_port [Integer|None] The port to write. Pass None to make this
    #   a read only stream.
    # self.param read_port [Integer|None] The port to read. Pass None to make this
    #   a write only stream.
    # self.param write_timeout [Float] Seconds to wait before aborting writes
    # self.param read_timeout [Float|None] Seconds to wait before aborting reads.
    #   Pass None to block until the read is complete.
    # self.param connect_timeout [Float|None] Seconds to wait before aborting connect.
    #   Pass None to block until the connection is complete.
    def __init__(
        self,
        hostname,
        write_port,
        read_port,
        write_timeout,
        read_timeout,
        connect_timeout=5.0,
    ):
        try:
            socket.gethostbyname(hostname)
        except socket.gaierror as error:
            raise RuntimeError(f"Invalid hostname {hostname}") from error
        self.hostname = hostname
        if str(hostname).upper() == "LOCALHOST":
            self.hostname = "127.0.0.1"
        self.write_port = ConfigParser.handle_none(write_port)
        if self.write_port:
            self.write_port = int(write_port)
        self.read_port = ConfigParser.handle_none(read_port)
        if self.read_port:
            self.read_port = int(read_port)

        write_socket = None
        if self.write_port:
            write_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            write_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            write_socket.setblocking(False)
        read_socket = None
        if self.read_port:
            if self.write_port != self.read_port:
                read_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
                read_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                read_socket.setblocking(False)
            else:
                read_socket = write_socket

        self.connect_timeout = ConfigParser.handle_none(connect_timeout)
        if self.connect_timeout:
            self.connect_timeout = float(connect_timeout)

        super().__init__(write_socket, read_socket, write_timeout, read_timeout)

    # Connect the socket(s)
    def connect(self):
        if self.write_socket:
            self._connect(self.write_socket, self.hostname, self.write_port)
        if self.read_socket and self.read_socket != self.write_socket:
            self._connect(self.read_socket, self.hostname, self.read_port)
        super().connect()

    def _connect(self, socket, hostname, port):
        while True:
            try:
                socket.connect((hostname, port))
            except BlockingIOError:
                # select.select([], [socket], [], self.connect_timeout)
                # This is not an error condition
                continue
            except OSError as error:
                if error.errno == errno.EINPROGRESS:
                    continue
                if error.errno == errno.EISCONN or error.errno == errno.EALREADY:
                    break
                else:
                    raise error

    # except:
    #   try:
    #     _, sockets, _ = IO.select(None, [socket], None, self.connect_timeout) # wait 3-way handshake completion
    #   except IOError, Errno='ENOTSOCK':
    #     raise "Connect canceled"
    #   if sockets and !sockets.empty?:
    #     try:
    #       socket.connect_nonblock(addr) # check connection failure
    #     except IOError, Errno='ENOTSOCK':
    #       raise "Connect canceled"
    #     except Errno='EINPROGRESS':
    #       retry
    #     except Errno='EISCONN', Errno='EALREADY':
    #   else:
    #     raise "Connect timeout"
    # except IOError, Errno='ENOTSOCK':
    #   raise "Connect canceled"



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.models.router_model import RouterModel
from openc3.models.router_status_model import RouterStatusModel
from openc3.topics.router_topic import RouterTopic

WHITELIST.extend(
    [
        "get_router",
        "get_router_names",
        "connect_router",
        "disconnect_router",
        "start_raw_logging_router",
        "stop_raw_logging_router",
        "get_all_router_info",
        "router_cmd",
        "router_protocol_cmd",
    ]
)


# Get information about an router
#
# @since 5.0.0
# @param router_name [String] Router name
# @return [Hash] Hash of all the router information
def get_router(router_name, scope=OPENC3_SCOPE):
    authorize(permission="system", router_name=router_name, scope=scope)
    router = RouterModel.get(name=router_name, scope=scope)
    if not router:
        raise RuntimeError(f"Router '{router_name}' does not exist")
    return router | RouterStatusModel.get(name=router_name, scope=scope)


# @return [Array<String>] All the router names
def get_router_names(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return RouterModel.names(scope=scope)


# Connects an router and starts its telemetry gathering thread
#
# @param router_name [String] The name of the router
# @param router_params [Array] Optional parameters to pass to the router
def connect_router(router_name, *router_params, scope=OPENC3_SCOPE):
    authorize(permission="system_set", router_name=router_name, scope=scope)
    RouterTopic.connect_router(router_name, *router_params, scope=scope)


# Disconnects from an router and kills its telemetry gathering thread
#
# @param router_name [String] The name of the router
def disconnect_router(router_name, scope=OPENC3_SCOPE):
    authorize(permission="system_set", router_name=router_name, scope=scope)
    RouterTopic.disconnect_router(router_name, scope=scope)


# Starts raw logging for an router
#
# @param router_name [String] The name of the router
def start_raw_logging_router(router_name="ALL", scope=OPENC3_SCOPE):
    authorize(permission="system_set", router_name=router_name, scope=scope)
    if router_name == "ALL":
        for router_name in get_router_names():
            RouterTopic.start_raw_logging(router_name, scope=scope)
    else:
        RouterTopic.start_raw_logging(router_name, scope=scope)


# Stop raw logging for an router
#
# @param router_name [String] The name of the router
def stop_raw_logging_router(router_name="ALL", scope=OPENC3_SCOPE):
    authorize(permission="system_set", router_name=router_name, scope=scope)
    if router_name == "ALL":
        for router_name in get_router_names():
            RouterTopic.stop_raw_logging(router_name, scope=scope)
    else:
        RouterTopic.stop_raw_logging(router_name, scope=scope)


# Consolidate all router info into a single API call
#
# @return [Array<Array<String, Numeric, Numeric, Numeric, Numeric, Numeric,
#   Numeric, Numeric>>] Array of Arrays containing \[name, state, num clients,
#   TX queue size, RX queue size, TX bytes, RX bytes, Command count,
#   Telemetry count] for all routers
def get_all_router_info(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    info = []
    for _, router in RouterStatusModel.all(scope=scope).items():
        info.append(
            [
                router["name"],
                router["state"],
                router["clients"],
                router["txsize"],
                router["rxsize"],
                router["txbytes"],
                router["rxbytes"],
                router["rxcnt"],
                router["txcnt"],
            ]
        )
    return info


def router_cmd(router_name, cmd_name, *cmd_params, scope=OPENC3_SCOPE):
    authorize(permission="system_set", router_name=router_name, scope=scope)
    RouterTopic.router_cmd(router_name, cmd_name, *cmd_params, scope=scope)


def router_protocol_cmd(
    router_name,
    cmd_name,
    *cmd_params,
    read_write="READ_WRITE",
    index=-1,
    scope=OPENC3_SCOPE,
):
    authorize(permission="system_set", router_name=router_name, scope=scope)
    RouterTopic.protocol_cmd(
        router_name,
        cmd_name,
        *cmd_params,
        read_write=read_write,
        index=index,
        scope=scope,
    )



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from openc3.api import *
from openc3.config.config_parser import ConfigParser


# Base class for all OpenC3 protocols which defines a framework which must be
# implemented by a subclass.
class Protocol:
    # self.param allow_empty_data [True/False/None] Whether or not this protocol will allow an empty string
    # to be passed down to later Protocols (instead of returning 'STOP'). Can be True, False, or None, where
    # None is interpreted as True if not the Protocol is the last Protocol of the chain.
    def __init__(self, allow_empty_data=None):
        self.interface = None
        self.allow_empty_data = ConfigParser.handle_true_false_none(allow_empty_data)
        self.reset()

    def reset(self):
        self.extra = None

    def connect_reset(self):
        self.reset()

    def disconnect_reset(self):
        self.reset()

    # Ensure we have some data in match this is the only protocol:
    def read_data(self, data, extra=None):
        if len(data) <= 0:
            if self.allow_empty_data is None:
                if self.interface and self.interface.read_protocols[-1] == self:
                    # Last read interface in chain with auto self.allow_empty_data
                    return ("STOP", extra)
            elif self.allow_empty_data:
                # Don't self.allow_empty_data means STOP
                return ("STOP", extra)
        return (data, extra)

    def read_packet(self, packet):
        return packet

    def write_packet(self, packet):
        return packet

    def write_data(self, data, extra=None):
        return (data, extra)

    def post_write_interface(self, packet, data, extra=None):
        return (packet, data, extra)

    def protocol_cmd(self, cmd_name, *cmd_args):
        # Default do nothing - Implemented by subclasses
        return False



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from datetime import datetime
import openc3.script
from openc3.environment import OPENC3_SCOPE
from openc3.top_level import HazardousError, CriticalCmdError
from openc3.utilities.extract import *
from openc3.packets.packet import Packet


def cmd(*args, **kwargs):
    """Send a command to the specified target
    Usage:
      cmd(target_name, cmd_name, cmd_params = {})
    or
      cmd('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd", "cmd_no_hazardous_check", *args, **kwargs)


def cmd_raw(*args, **kwargs):
    """Send a command to the specified target without running conversions
    Usage:
      cmd_raw(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw", "cmd_raw_no_hazardous_check", *args, **kwargs)


def cmd_no_range_check(*args, **kwargs):
    """Send a command to the specified target without range checking parameters
    Usage:
      cmd_no_range_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_no_range_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_no_range_check", "cmd_no_checks", *args, **kwargs)


def cmd_raw_no_range_check(*args, **kwargs):
    """Send a command to the specified target without range checking parameters or running conversions
    Usage:
      cmd_raw_no_range_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw_no_range_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw_no_range_check", "cmd_raw_no_checks", *args, **kwargs)


def cmd_no_hazardous_check(*args, **kwargs):
    """Send a command to the specified target without hazardous checks
    Usage:
      cmd_no_hazardous_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_no_hazardous_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_no_hazardous_check", None, *args, **kwargs)


def cmd_raw_no_hazardous_check(*args, **kwargs):
    """Send a command to the specified target without hazardous checks or running conversions
    Usage:
      cmd_raw_no_hazardous_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw_no_hazardous_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw_no_hazardous_check", None, *args, **kwargs)


def cmd_no_checks(*args, **kwargs):
    """Send a command to the specified target without range checking or hazardous checks
    Usage:
      cmd_no_checks(target_name, cmd_name, cmd_params = {})
    or
      cmd_no_checks('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_no_checks", None, *args, **kwargs)


def cmd_raw_no_checks(*args, **kwargs):
    """Send a command to the specified target without range checking or hazardous checks or running conversions
    Usage:
      cmd_raw_no_checks(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw_no_checks('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw_no_checks", None, *args, **kwargs)


def build_cmd(*args, **kwargs):
    """Builds a command binary
    Accepts two different calling styles:
    build_command("TGT CMD with PARAM1 val, PARAM2 val")
    build_command('TGT','CMD',{'PARAM1'=>val,'PARAM2'=>val})"""
    extract_string_kwargs_to_args(args, kwargs)
    return getattr(openc3.script.API_SERVER, "build_cmd")(*args)


# build_command is DEPRECATED
build_command = build_cmd


def get_cmd_hazardous(*args, **kwargs):
    """Returns whether a command is hazardous (true or false)"""
    extract_string_kwargs_to_args(args, kwargs)
    return getattr(openc3.script.API_SERVER, "get_cmd_hazardous")(*args)


# Returns the time the most recent command was sent
def get_cmd_time(target_name=None, command_name=None, scope=OPENC3_SCOPE):
    results = getattr(openc3.script.API_SERVER, "get_cmd_time")(target_name, command_name, scope=scope)
    results = list(results)
    if results[2] and results[3]:
        results[2] = datetime.fromtimestamp(results[2] + results[3] / 1000000)
        results.pop(3)
    return results


# Format the command like it appears in a script
def _cmd_string(target_name, cmd_name, cmd_params, raw):
    output_string = ""
    if openc3.script.DISCONNECT:
        output_string += "openc3.script.DISCONNECT: "
    if raw:
        output_string += 'cmd_raw("'
    else:
        output_string += 'cmd("'
    output_string += target_name + " " + cmd_name
    if not cmd_params:
        output_string += '")'
    else:
        params = []
        for key, value in cmd_params.items():
            if key in Packet.RESERVED_ITEM_NAMES:
                continue
            if isinstance(value, str):
                if not value.isascii():
                    value = "BINARY"
                elif len(value) > 256:
                    value = value[:255] + "...'"
                value = value.replace('"', "'")
            elif isinstance(value, list):
                value = str(value)
            params.append(f"{key} {value}")
        params = ", ".join(params)
        output_string += " with " + params + '")'
    return output_string


def _log_cmd(target_name, cmd_name, cmd_params, raw, no_range, no_hazardous):
    """Log any warnings about disabling checks and log the command itself
    NOTE: This is a helper method and should not be called directly"""
    if no_range:
        print(f"WARN: Command {target_name} {cmd_name} being sent ignoring range checks")
    if no_hazardous:
        print(f"WARN: Command {target_name} {cmd_name} being sent ignoring hazardous warnings")
    print(_cmd_string(target_name, cmd_name, cmd_params, raw))


def _cmd_disconnect(cmd, raw, no_range, no_hazardous, *args, scope):
    match len(args):
        case 1:
            target_name, cmd_name, cmd_params = extract_fields_from_cmd_text(args[0])
        case 2 | 3:
            target_name = args[0]
            cmd_name = args[1]
            if len(args) == 2:
                cmd_params = {}
            else:
                cmd_params = args[2]
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {cmd}()")

    # Get the command and validate the parameters
    command = openc3.script.API_SERVER.get_cmd(target_name, cmd_name, scope=scope)
    if cmd_params:
        for param_name in cmd_params.keys():
            found = False
            if "items" in command:
                for item in command["items"]:
                    if item["name"] == param_name:
                        found = item
                        break
            if not found:
                raise RuntimeError(f"Packet item '{target_name} {cmd_name} {param_name}' does not exist")
    _log_cmd(target_name, cmd_name, cmd_params, raw, no_range, no_hazardous)


def _cmd(
    cmd,
    cmd_no_hazardous,
    *args,
    timeout=None,
    log_message=None,
    validate=True,
    scope=OPENC3_SCOPE,
):
    """Send the command and log the results
    # This method signature has to include the keyword params present in cmd_api.py cmd_implementation()
    NOTE: This is a helper method and should not be called directly"""

    raw = "raw" in cmd
    no_range = "no_range" in cmd or "no_checks" in cmd
    no_hazardous = "no_hazardous" in cmd or "no_checks" in cmd

    if openc3.script.DISCONNECT:
        _cmd_disconnect(cmd, raw, no_range, no_hazardous, *args, scope=scope)
    else:
        try:
            try:
                target_name, cmd_name, cmd_params = getattr(openc3.script.API_SERVER, cmd)(
                    *args, timeout=timeout, log_message=log_message, validate=validate, scope=scope
                )
                if log_message is None or log_message:
                    _log_cmd(target_name, cmd_name, cmd_params, raw, no_range, no_hazardous)
            except HazardousError as error:
                # Need to reimport here to pick up changes from running_script
                from openc3.script import prompt_for_hazardous

                # This opens a prompt at which point they can cancel and stop the script
                # or say Yes and send the command. Thus we don't care about the return value.
                prompt_for_hazardous(
                    error.target_name,
                    error.cmd_name,
                    error.hazardous_description,
                )
                target_name, cmd_name, cmd_params = getattr(openc3.script.API_SERVER, cmd_no_hazardous)(
                    *args, timeout=timeout, log_message=log_message, validate=validate, scope=scope
                )
                if log_message is None or log_message:
                    _log_cmd(target_name, cmd_name, cmd_params, raw, no_range, no_hazardous)
        except CriticalCmdError as error:
            # Need to reimport here to pick up changes from running_script
            from openc3.script import prompt_for_critical_cmd

            # This should not return until the critical command has been approved
            prompt_for_critical_cmd(
                error.uuid,
                error.username,
                error.target_name,
                error.cmd_name,
                error.cmd_params,
                error.cmd_string,
            )
            if log_message is None or log_message:
                _log_cmd(error.target_name, error.cmd_name, error.cmd_params, raw, no_range, no_hazardous)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.conversions.conversion import Conversion


class BitReverseConversion(Conversion):
    def __init__(self, converted_type, converted_bit_size):
        super().__init__()
        self.converted_type = str(converted_type).upper()
        self.converted_bit_size = int(converted_bit_size)
        if self.converted_type == "FLOAT":
            raise RuntimeError("Float Bit Reverse Not Yet Supported")
        self.params = [self.converted_type, self.converted_bit_size]

    # Perform the conversion on the value.
    #
    # @param value [Object] The value to convert
    # @param packet [Packet] The packet which contains the value. This can
    #   be useful to reach into the packet and use other values in the
    #   conversion.
    # @param buffer [String] The packet buffer
    # @return The converted value
    def call(self, value, _packet, _buffer):
        b = "{:0{width}b}".format(value, width=self.converted_bit_size)
        return int(b[::-1], 2)

    # @return [String] The conversion class
    def __str__(self):
        return f"BitReverseConversion {self.converted_type} {self.converted_bit_size}"

    # @param read_or_write [String] Either 'READ' or 'WRITE'
    # @return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        return f"{read_or_write}_CONVERSION openc3/conversions/bit_reverse_conversion.py {self.converted_type} {self.converted_bit_size}\n"



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import select
import socket
import threading
import multiprocessing
from openc3.streams.stream import Stream
from openc3.config.config_parser import ConfigParser
from openc3.utilities.logger import Logger
from openc3.top_level import close_socket


class TcpipSocketStream(Stream):
    # self.param write_socket [Socket] Socket to write
    # self.param read_socket [Socket] Socket to read
    # self.param write_timeout [Float] Seconds to wait before aborting writes
    # self.param read_timeout [Float|None] Seconds to wait before aborting reads.
    #   Pass None to block until the read is complete.
    def __init__(self, write_socket, read_socket, write_timeout, read_timeout):
        super().__init__()
        self.write_socket = write_socket
        self.read_socket = read_socket
        self.write_timeout = ConfigParser.handle_none(write_timeout)
        if self.write_timeout:
            self.write_timeout = float(write_timeout)
        else:
            Logger.warn("Warning: To avoid interface lock, write_timeout can not be None. Setting to 10 seconds.")
            self.write_timeout = 10.0
        self.read_timeout = ConfigParser.handle_none(read_timeout)
        if self.read_timeout:
            self.read_timeout = float(read_timeout)

        # Mutex on write is needed to protect from commands coming in from more
        # than one tool
        self.write_mutex = threading.Lock()
        self.pipe_reader, self.pipe_writer = multiprocessing.Pipe()
        self._connected = False

    # self.return [String] Returns a binary string of data from the socket
    def read(self):
        if not self.read_socket:
            raise RuntimeError("Attempt to read from write only stream")

        data = ""
        # No read mutex is needed because reads happen serially
        while True:  # Loop until we get some data
            try:
                data = self.read_socket.recv(4096, socket.MSG_DONTWAIT)
            # Non-blocking sockets return an errno EAGAIN or EWOULDBLOCK
            # if there is no data available
            except socket.error as error:
                if error.errno == socket.EAGAIN or error.errno == socket.EWOULDBLOCK:
                    # If select returns something it means the socket is now available for
                    # reading so retry the read. If it returns empty list it means we timed out.
                    # If the pipe is present that means we closed the socket
                    readable, _, _ = select.select(
                        [self.read_socket, self.pipe_reader],
                        [],
                        [],
                        self.read_timeout,
                    )
                    if readable and self.pipe_reader in readable:
                        data = ""
                    else:
                        continue
            break
        return data

    # self.param data [String] A binary string of data to write to the socket
    def write(self, data):
        if not self.write_socket:
            raise RuntimeError("Attempt to write to read only stream")

        with self.write_mutex:
            num_bytes_to_send = len(data)
            total_bytes_sent = 0
            bytes_sent = 0
            data_to_send = data

            while True:
                try:
                    bytes_sent = self.write_socket.send(data_to_send, socket.MSG_DONTWAIT)
                # Non-blocking sockets return an errno EAGAIN or EWOULDBLOCK
                # if the write would block
                except socket.error as error:
                    if error.errno == socket.EAGAIN or error.errno == socket.EWOULDBLOCK:
                        # Wait for the socket to be ready for writing or for the timeout
                        _, writeable, _ = select.select([], [self.write_socket], [], self.write_timeout)
                        # If select returns something it means the socket is now available for
                        # writing so retry the write. If it returns None it means we timed out.
                        if writeable:
                            continue
                        else:
                            raise RuntimeError("Write Timeout")
                total_bytes_sent += bytes_sent
                if total_bytes_sent >= num_bytes_to_send:
                    break

                data_to_send = data[total_bytes_sent:]

    # Connect the stream
    def connect(self):
        # If called directly this class is acting as a server and does not need to connect the sockets
        self._connected = True

    def connected(self):
        return self._connected

    # Disconnect by closing the sockets
    def disconnect(self):
        if not self._connected:
            return
        close_socket(self.write_socket)
        close_socket(self.read_socket)
        self.pipe_writer.send(".")
        self._connected = False



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

__title__ = "openc3"
__description__ = "Python Support for OpenC3 COSMOS"
__url__ = "https://github.com/OpenC3/cosmos"
__version__ = "6.4.1"



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.models.interface_model import InterfaceModel
from openc3.models.interface_status_model import InterfaceStatusModel
from openc3.topics.interface_topic import InterfaceTopic
from openc3.utilities.logger import Logger

WHITELIST.extend(
    [
        "get_interface",
        "get_interface_names",
        "connect_interface",
        "disconnect_interface",
        "start_raw_logging_interface",
        "stop_raw_logging_interface",
        "get_all_interface_info",
        "map_target_to_interface",
        "interface_cmd",
        "interface_protocol_cmd",
    ]
)


# Get information about an interface
#
# @since 5.0.0
# @param interface_name [String] Interface name
# @return [Hash] Hash of all the interface information
def get_interface(interface_name, scope=OPENC3_SCOPE):
    authorize(permission="system", interface_name=interface_name, scope=scope)
    interface = InterfaceModel.get(name=interface_name, scope=scope)
    if not interface:
        raise RuntimeError(f"Interface '{interface_name}' does not exist")
    return interface | InterfaceStatusModel.get(name=interface_name, scope=scope)


# @return [Array<String>] All the interface names
def get_interface_names(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return InterfaceModel.names(scope=scope)


# Connects an interface and starts its telemetry gathering thread
#
# @param interface_name [String] The name of the interface
# @param interface_params [Array] Optional parameters to pass to the interface
def connect_interface(interface_name, *interface_params, scope=OPENC3_SCOPE):
    authorize(permission="system_set", interface_name=interface_name, scope=scope)
    InterfaceTopic.connect_interface(interface_name, *interface_params, scope=scope)


# Disconnects from an interface and kills its telemetry gathering thread
#
# @param interface_name [String] The name of the interface
def disconnect_interface(interface_name, scope=OPENC3_SCOPE):
    authorize(permission="system_set", interface_name=interface_name, scope=scope)
    InterfaceTopic.disconnect_interface(interface_name, scope=scope)


# Starts raw logging for an interface
#
# @param interface_name [String] The name of the interface
def start_raw_logging_interface(interface_name="ALL", scope=OPENC3_SCOPE):
    authorize(permission="system_set", interface_name=interface_name, scope=scope)
    if interface_name == "ALL":
        for interface_name in get_interface_names():
            InterfaceTopic.start_raw_logging(interface_name, scope=scope)
    else:
        InterfaceTopic.start_raw_logging(interface_name, scope=scope)


# Stop raw logging for an interface
#
# @param interface_name [String] The name of the interface
def stop_raw_logging_interface(interface_name="ALL", scope=OPENC3_SCOPE):
    authorize(permission="system_set", interface_name=interface_name, scope=scope)
    if interface_name == "ALL":
        for interface_name in get_interface_names():
            InterfaceTopic.stop_raw_logging(interface_name, scope=scope)
    else:
        InterfaceTopic.stop_raw_logging(interface_name, scope=scope)


# Get information about all interfaces
#
# @return [Array<Array<String, Numeric, Numeric, Numeric, Numeric, Numeric,
#   Numeric, Numeric>>] Array of Arrays containing \[name, state, num clients,
#   TX queue size, RX queue size, TX bytes, RX bytes, Command count,
#   Telemetry count] for all interfaces
def get_all_interface_info(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    info = []
    for int_name, int in InterfaceStatusModel.all(scope=scope).items():
        # Get the interface configuration to access disable_disconnect
        interface_model = InterfaceModel.get(name=int_name, scope=scope)
        if interface_model and interface_model['disable_disconnect']:
            disable_disconnect = True
        else:
            disable_disconnect = False
        info.append(
            [
                int["name"],
                int["state"],
                int["clients"],
                int["txsize"],
                int["rxsize"],
                int["txbytes"],
                int["rxbytes"],
                int["txcnt"],
                int["rxcnt"],
                disable_disconnect,
            ]
        )
    return info


# Associates a target and all its commands and telemetry with a particular
# interface. All the commands will go out over and telemetry be received
# from that interface.
#
# @param target_name [String/Array] The name of the target(s)
# @param interface_name (see #connect_interface)
def map_target_to_interface(
    target_name,
    interface_name,
    cmd_only=False,
    tlm_only=False,
    unmap_old=True,
    scope=OPENC3_SCOPE,
):
    authorize(permission="system_set", interface_name=interface_name, scope=scope)
    new_interface = InterfaceModel.get_model(name=interface_name, scope=scope)
    if isinstance(target_name, list):
        target_names = target_name
    else:
        target_names = [target_name]
    for name in target_names:
        new_interface.map_target(name, cmd_only=cmd_only, tlm_only=tlm_only, unmap_old=unmap_old)
        Logger.info(f"Target {name} mapped to Interface {interface_name}", scope=scope)


def interface_cmd(interface_name, cmd_name, *cmd_params, scope=OPENC3_SCOPE):
    authorize(permission="system_set", interface_name=interface_name, scope=scope)
    InterfaceTopic.interface_cmd(interface_name, cmd_name, *cmd_params, scope=scope)


def interface_protocol_cmd(
    interface_name,
    cmd_name,
    *cmd_params,
    read_write="READ_WRITE",
    index=-1,
    scope=OPENC3_SCOPE,
):
    authorize(permission="system_set", interface_name=interface_name, scope=scope)
    InterfaceTopic.protocol_cmd(
        interface_name,
        cmd_name,
        *cmd_params,
        read_write=read_write,
        index=index,
        scope=scope,
    )



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import redis
from redis.exceptions import TimeoutError
from openc3.utilities.connection_pool import ConnectionPool
from contextlib import contextmanager
import threading
from openc3.environment import *

if OPENC3_REDIS_CLUSTER:
    openc3_redis_cluster = True
else:
    openc3_redis_cluster = False


class StoreConnectionPool(ConnectionPool):
    @contextmanager
    def pipelined(self):
        if openc3_redis_cluster:
            yield  # TODO: Update keys to support pipelining in cluster
        else:
            with self.get() as redis:
                pipeline = redis.pipeline(transaction=False)
                thread_id = threading.get_native_id()
                self.pipelines[thread_id] = pipeline
                try:
                    yield
                finally:
                    pipeline.execute()
                    self.pipelines[thread_id] = None

    @contextmanager
    def get(self):
        thread_id = threading.get_native_id()
        if thread_id not in self.pipelines:
            self.pipelines[thread_id] = None
        pipeline = self.pipelines[thread_id]
        if pipeline:
            yield pipeline
        else:
            item = None
            with self.lock:
                if not self.pool.empty():
                    item = self.pool.get(False)
                elif self.count < self.pool_size:
                    item = self.ctor()
                    self.count += 1
                else:
                    item = self.pool.get()
            try:
                yield item
            finally:
                self.pool.put(item)


class StoreMeta(type):
    def __getattribute__(cls, func):
        if func == "instance" or func == "instance_mutex" or func == "my_instance":
            return super().__getattribute__(func)

        # Handle dunder methods to support help() and other introspection
        if func.startswith("__") and func.endswith("__"):
            return super().__getattribute__(func)

        def method(*args, **kw_args):
            return getattr(cls.instance(), func)(*args, **kw_args)

        return method


class Store(metaclass=StoreMeta):
    # Variable that holds the singleton instance
    my_instance = None

    # Mutex used to ensure that only one instance is created
    instance_mutex = threading.Lock()

    # Get the singleton instance
    @classmethod
    def instance(cls, pool_size=100):
        if cls.my_instance:
            return cls.my_instance

        with cls.instance_mutex:
            cls.my_instance = cls(pool_size)
            return cls.my_instance

    # Delegate all unknown methods to redis through the @redis_pool
    def __getattr__(self, func):
        with self.redis_pool.get() as redis:

            def method(*args, **kwargs):
                return getattr(redis, func)(*args, **kwargs)

            return method

    def __init__(self, pool_size=10):
        self.redis_host = OPENC3_REDIS_HOSTNAME
        self.redis_port = OPENC3_REDIS_PORT
        self.redis_pool = StoreConnectionPool(self.build_redis, pool_size)
        self.topic_offsets = {}
        self.pipelines = {}

    if not openc3_redis_cluster:

        def build_redis(self):
            # NOTE: We can't use decode_response because it tries to decode the binary
            # packet buffer which does not work. Thus strings come back as bytes like
            # b"target_name" and we decode them using b"target_name".decode()
            return redis.Redis(
                host=self.redis_host,
                port=self.redis_port,
                username=OPENC3_REDIS_USERNAME,
                password=OPENC3_REDIS_PASSWORD,
            )

    ###########################################################################
    # Stream APIs
    ###########################################################################

    def get_oldest_message(self, topic):
        with self.redis_pool.get() as redis:
            result = redis.xrange(topic, count=1)
            if result and len(result) > 0:
                return result[0]
            else:
                return None

    def get_newest_message(self, topic):
        with self.redis_pool.get() as redis:
            # Default in xrevrange is range end '+', start '-' which means get all
            # elements from higher ID to lower ID and since we're limiting to 1
            # we get the last element. See https://redis.io/commands/xrevrange.
            result = redis.xrevrange(topic, count=1)
            if result and len(result) > 0:
                first = list(result[0])
                first[0] = first[0].decode()
                return first
            else:
                return (None, None)

    def get_last_offset(self, topic):
        with self.redis_pool.get() as redis:
            result = redis.xrevrange(topic, count=1)
            if result and result[0] and result[0][0]:
                return result[0][0].decode()
            else:
                return "0-0"

    def update_topic_offsets(self, topics):
        offsets = []
        for topic in topics:
            # Normally we will just be grabbing the topic offset
            # this allows xread to get everything past this point
            thread_id = threading.get_native_id()
            if thread_id not in self.topic_offsets:
                self.topic_offsets[thread_id] = {}
            topic_offsets = self.topic_offsets[thread_id]
            last_id = topic_offsets.get(topic)
            if last_id:
                offsets.append(last_id)
            else:
                # If there is no topic offset this is the first call.
                # Get the last offset ID so we'll start getting everything from now on
                offsets.append(self.get_last_offset(topic))
                topic_offsets[topic] = offsets[-1]
        return offsets

    if not openc3_redis_cluster:

        def read_topics(self, topics, offsets=None, timeout_ms=1000, count=None):
            if len(topics) == 0:
                return {}
            thread_id = threading.get_native_id()
            if thread_id not in self.topic_offsets:
                self.topic_offsets[thread_id] = {}
            topic_offsets = self.topic_offsets[thread_id]
            try:
                with self.redis_pool.get() as redis:
                    if not offsets:
                        offsets = self.update_topic_offsets(topics)
                    streams = {}
                    index = 0
                    for topic in topics:
                        streams[topic] = offsets[index]
                        index += 1
                    result = redis.xread(streams, block=timeout_ms, count=count)
                    if result and len(result) > 0:
                        for topic, messages in result:
                            for msg_id, msg_hash in messages:
                                if isinstance(topic, bytes):
                                    topic = topic.decode()
                                if isinstance(msg_id, bytes):
                                    msg_id = msg_id.decode()
                                topic_offsets[topic] = msg_id
                                yield topic, msg_id, msg_hash, redis
                    return result
            except TimeoutError:
                # Should return an empty hash not array - xread returns a hash
                return {}

    # Add new entry to the redis stream.
    # > https://www.rubydoc.info/github/redis/redis-rb/Redis:xadd
    #
    # @example Without options
    #   store.write_topic('MANGO__TOPIC', {'message' => 'something'})
    # @example With options
    #   store.write_topic('MANGO__TOPIC', {'message' => 'something'}, id: '0-0', maxlen: 1000, approximate: 'true')
    #
    # @param topic [String] the stream / topic
    # @param msg_hash [Hash]   one or multiple field-value pairs
    #
    # @option opts [String]  :id          the entry id, default value is `*`, it means auto generation,
    #   if `nil` id is passed it will be changed to `*`
    # @option opts [Integer] :maxlen      max length of entries, default value is `nil`, it means will grow forever
    # @option opts [String] :approximate whether to add `~` modifier of maxlen or not, default value is 'true'
    #
    # @return [String] the entry id
    def write_topic(self, topic, msg_hash, id="*", maxlen=None, approximate=True):
        if not id:
            id = "*"
        with self.redis_pool.get() as redis:
            return redis.xadd(topic, msg_hash, id=id, maxlen=maxlen, approximate=approximate)

    # Trims older entries of the redis stream if needed.
    # > https://www.rubydoc.info/github/redis/redis-rb/Redis:xtrim
    #
    # @example Without options
    #   store.trim_topic('MANGO__TOPIC', 1000)
    # @example With options
    #   store.trim_topic('MANGO__TOPIC', 1000, approximate: 'true', limit: 0)
    #
    # @param topic  [String]  the stream key
    # @param minid  [Integer] Id to throw away data up to
    # @param approximate [Boolean] whether to add `~` modifier of maxlen or not
    # @param limit  [Boolean] number of items to return from the call
    #
    # @return [Integer] the number of entries actually deleted
    def trim_topic(self, topic, minid, approximate=True, limit=0):
        with self.redis_pool.get() as redis:
            return redis.xtrim(name=topic, minid=minid, approximate=approximate, limit=limit)


class EphemeralStore(Store):
    # Variable that holds the singleton instance
    my_instance = None

    def __init__(self, pool_size=10):
        super().__init__(pool_size)
        self.redis_host = OPENC3_REDIS_EPHEMERAL_HOSTNAME
        self.redis_port = OPENC3_REDIS_EPHEMERAL_PORT
        self.redis_pool = StoreConnectionPool(self.build_redis, pool_size)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.utilities.extract import *
import openc3.script
from openc3.environment import OPENC3_SCOPE


def _script_response_error(response, message, scope=OPENC3_SCOPE):
    if response:
        raise RuntimeError(f"{message} ({response.status_code}): {response.text}")
    else:
        raise RuntimeError(f"{message}: No Response")


def script_list(scope=OPENC3_SCOPE):
    endpoint = "/script-api/scripts"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("get", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, "Script list request failed", scope=scope)
    else:
        scripts = json.loads(response.text)
        # Remove the '*' from the script names
        return [script.rstrip("*") for script in scripts]


def script_syntax_check(script, scope=OPENC3_SCOPE):
    endpoint = "/script-api/scripts/temp.py/syntax"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("post", endpoint, json=False, data=script, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, "Script syntax check request failed", scope=scope)
    else:
        result = json.loads(response.text)
        if result.get("title") == "Syntax Check Successful":
            result["success"] = True
        else:
            result["success"] = False
        return result


def script_body(filename, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/scripts/{filename}"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("get", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, f"Failed to get {filename}", scope=scope)
    else:
        result = json.loads(response.text)
        return result.get("contents")


def script_run(filename, disconnect=False, environment=None, scope=OPENC3_SCOPE):
    if disconnect:
        endpoint = f"/script-api/scripts/{filename}/run/disconnect"
    else:
        endpoint = f"/script-api/scripts/{filename}/run"

    # Encode the environment hash into an array of key values
    if environment and not len(environment) == 0:
        env_data = []
        for key, value in environment.items():
            env_data.append({"key": key, "value": value})
    else:
        env_data = []
    # NOTE: json: true causes json_api_object to JSON generate and set the Content-Type to json
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request(
        "post", endpoint, json=True, data={"environment": env_data}, scope=scope
    )
    if not response or response.status_code != 200:
        _script_response_error(response, f"Failed to run {filename}", scope=scope)
    else:
        script_id = int(response.text)
        return script_id


def script_delete(filename, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/scripts/{filename}/delete"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("post", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, f"Failed to delete {filename}", scope=scope)
    else:
        return True


def script_lock(filename, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/scripts/{filename}/lock"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("post", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, f"Failed to lock {filename}", scope=scope)
    else:
        return True


def script_unlock(filename, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/scripts/{filename}/unlock"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("post", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, f"Failed to unlock {filename}", scope=scope)
    else:
        return True


def script_instrumented(script, scope=OPENC3_SCOPE):
    endpoint = "/script-api/scripts/temp.py/instrumented"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("post", endpoint, json=False, data=script, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, "Script instrumented request failed", scope=scope)
    else:
        result = json.loads(response.text)
        if result.get("title") == "Instrumented Script":
            parsed = json.loads(result.get("description"))
            return "\n".join(parsed)
        else:
            raise result


def script_create(filename, script, breakpoints=[], scope=OPENC3_SCOPE):
    endpoint = f"/script-api/scripts/{filename}"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request(
        "post",
        endpoint,
        json=True,
        data={"text": script, "breakpoints": breakpoints},
        scope=scope,
    )
    if not response or response.status_code != 200:
        _script_response_error(response, "Script create request failed", scope=scope)
    else:
        return json.loads(response.text)


def script_delete_all_breakpoints(scope=OPENC3_SCOPE):
    endpoint = "/script-api/breakpoints/delete/all"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("delete", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, "Script delete all breakpoints failed", scope=scope)
    else:
        return True


def running_script_list(limit=10, offset=0, scope=OPENC3_SCOPE):
    endpoint = "/script-api/running-script"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("get", endpoint, query={"limit": limit, "offset": offset}, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, "Running script list request failed", scope=scope)
    else:
        return json.loads(response.text)['items']


def script_get(id, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/running-script/{id}"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("get", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, "Running script show request failed", scope=scope)
    else:
        return json.loads(response.text)
running_script_get = script_get  # Alias for compatibility


def _running_script_action(id, action_name, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/running-script/{id}/{action_name}"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("post", endpoint, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, f"Running script {action_name} request failed", scope=scope)
    else:
        return True


def running_script_stop(id, scope=OPENC3_SCOPE):
    _running_script_action(id, "stop", scope=scope)


def running_script_pause(id, scope=OPENC3_SCOPE):
    _running_script_action(id, "pause", scope=scope)


def running_script_retry(id, scope=OPENC3_SCOPE):
    _running_script_action(id, "retry", scope=scope)


def running_script_go(id, scope=OPENC3_SCOPE):
    _running_script_action(id, "go", scope=scope)


def running_script_step(id, scope=OPENC3_SCOPE):
    _running_script_action(id, "step", scope=scope)


def running_script_delete(id, scope=OPENC3_SCOPE):
    _running_script_action(id, "delete", scope=scope)


def running_script_backtrace(id, scope=OPENC3_SCOPE):
    _running_script_action(id, "backtrace", scope=scope)


def running_script_debug(id, debug_code, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/running-script/{id}/debug"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request(
        "post", endpoint, json=True, data={"args": debug_code}, scope=scope
    )
    if not response or response.status_code != 200:
        _script_response_error(response, "Running script debug request failed", scope=scope)
    else:
        return True


def running_script_prompt(id, method_name, answer, prompt_id, password=None, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/running-script/{id}/prompt"
    if password:
        response = openc3.script.SCRIPT_RUNNER_API_SERVER.request(
            "post",
            endpoint,
            json=True,
            data={
                "method": method_name,
                "answer": answer,
                "prompt_id": prompt_id,
                "password": password,
            },
            scope=scope,
        )
    else:
        response = openc3.script.SCRIPT_RUNNER_API_SERVER.request(
            "post",
            endpoint,
            json=True,
            data={"method": method_name, "answer": answer, "prompt_id": prompt_id},
            scope=scope,
        )
    if not response or response.status_code != 200:
        _script_response_error(response, "Running script prompt request failed", scope=scope)
    else:
        return True


def running_script_execute_while_paused(id, filename, line_no, end_line_no=None, scope=OPENC3_SCOPE):
    endpoint = f"/script-api/running-script/{id}/executewhilepaused"
    if end_line_no is not None:
        response = openc3.script.SCRIPT_RUNNER_API_SERVER.request(
            "post",
            endpoint,
            json=True,
            data={
                "args": [filename, line_no, end_line_no],
            },
            scope=scope,
        )
    else:
        response = openc3.script.SCRIPT_RUNNER_API_SERVER.request(
            "post",
            endpoint,
            json=True,
            data={"args": [filename, line_no]},
            scope=scope,
        )
    if not response or response.status_code != 200:
        _script_response_error(response, "Running script executewhilepaused request failed", scope=scope)
    else:
        return True


def completed_script_list(limit=10, offset=0, scope=OPENC3_SCOPE):
    endpoint = "/script-api/completed-scripts"
    response = openc3.script.SCRIPT_RUNNER_API_SERVER.request("get", endpoint, query={"limit": limit, "offset": offset}, scope=scope)
    if not response or response.status_code != 200:
        _script_response_error(response, "Completed script list request failed", scope=scope)
    else:
        return json.loads(response.text)['items']



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1963

import os
import sys
import time
import json
import threading
import uuid
from datetime import datetime, timezone
from openc3.microservices.microservice import Microservice
from openc3.microservices.interface_decom_common import handle_inject_tlm
from openc3.system.system import System
from openc3.models.scope_model import ScopeModel
from openc3.models.interface_model import InterfaceModel
from openc3.models.interface_status_model import InterfaceStatusModel
from openc3.models.router_model import RouterModel
from openc3.models.router_status_model import RouterStatusModel
from openc3.models.cvt_model import CvtModel
from openc3.models.target_model import TargetModel
from openc3.topics.topic import Topic
from openc3.topics.interface_topic import InterfaceTopic
from openc3.topics.router_topic import RouterTopic
from openc3.topics.command_topic import CommandTopic
from openc3.topics.command_decom_topic import CommandDecomTopic
from openc3.topics.telemetry_topic import TelemetryTopic
from openc3.config.config_parser import ConfigParser
from openc3.interfaces.interface import WriteRejectError
from openc3.utilities.logger import Logger
from openc3.utilities.sleeper import Sleeper
from openc3.utilities.time import from_nsec_from_epoch
from openc3.utilities.json import JsonDecoder
from openc3.utilities.store import Store
from openc3.utilities.store_queued import StoreQueued, EphemeralStoreQueued
from openc3.utilities.thread_manager import ThreadManager
from openc3.top_level import kill_thread

try:
    from openc3enterprise.models.critical_cmd_model import CriticalCmdModel
except ModuleNotFoundError:
    # Should never actually be used in COSMOS Core
    pass


class InterfaceCmdHandlerThread:
    def __init__(self, interface, tlm, logger=None, metric=None, scope=None):
        self.interface = interface
        self.tlm = tlm
        self.scope = scope
        scope_model = ScopeModel.get_model(name=scope)
        if scope_model is not None:
            self.critical_commanding = scope_model.critical_commanding
        else:
            self.critical_commanding = "OFF"
        self.logger = logger
        if not self.logger:
            self.logger = Logger()
        self.metric = metric
        self.count = 0
        self.directive_count = 0
        if self.metric is not None:
            self.metric.set(
                name="interface_directive_total",
                value=self.directive_count,
                type="counter",
            )
            self.metric.set(name="interface_cmd_total", value=self.count, type="counter")

    def start(self):
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.thread.start()
        ThreadManager.instance().register(self.thread, stop_object=self)
        return self.thread

    def stop(self):
        kill_thread(self, self.thread)

    def graceful_kill(self):
        InterfaceTopic.shutdown(self.interface, scope=self.scope)
        time.sleep(0.001)  # Allow other threads to run

    def run(self):
        # receive_commands does a while True and does not return
        InterfaceTopic.receive_commands(self.process_cmd, self.interface, self.scope)

    def process_cmd(self, topic, msg_id, msg_hash, redis):
        # OpenC3.with_context(msg_hash) do
        release_critical = False

        if msg_hash.get(b"shutdown"):
            return "Shutdown"

        msgid_seconds_from_epoch = int(msg_id.split("-")[0]) / 1000.0
        delta = time.time() - msgid_seconds_from_epoch
        if self.metric is not None:
            self.metric.set(
                name="interface_topic_delta_seconds",
                value=delta,
                type="gauge",
                unit="seconds",
                help="Delta time between data written to stream and interface cmd start",
            )

        if topic == "OPENC3__SYSTEM__EVENTS":
            msg = json.loads(msg_hash[b"event"].decode())
            if msg["type"] == "scope":
                if msg["name"] == self.scope:
                    self.critical_commanding = msg["critical_commanding"]
            return "SUCCESS"

        # Check for a raw write to the interface
        if "CMD}INTERFACE" in topic:
            self.directive_count += 1
            if self.metric is not None:
                self.metric.set(
                    name="interface_directive_total",
                    value=self.directive_count,
                    type="counter",
                )
            if msg_hash.get(b"shutdown"):
                self.logger.info(f"{self.interface.name}: Shutdown requested")
                InterfaceTopic.clear_topics(InterfaceTopic.topics(self.interface, scope=self.scope))
                return "SHUTDOWN"
            if msg_hash.get(b"connect"):
                self.logger.info(f"{self.interface.name}: Connect requested")
                params = []
                if msg_hash.get(b"params"):
                    params = json.loads(msg_hash[b"params"])
                self.interface = self.tlm.attempting(*params)
                return "SUCCESS"
            if msg_hash.get(b"disconnect"):
                self.logger.info(f"{self.interface.name}: Disconnect requested")
                self.tlm.disconnect(False)
                return "SUCCESS"
            if msg_hash.get(b"raw"):
                if self.interface.connected():
                    self.logger.info(f"{self.interface.name}: Write raw")
                    # A raw interface write results in an UNKNOWN packet
                    command = System.commands.packet("UNKNOWN", "UNKNOWN")
                    command.received_count = TargetModel.increment_command_count(command.target_name, command.packet_name, 1, scope=self.scope)
                    command = command.clone()
                    command.buffer = msg_hash[b"raw"]
                    command.received_time = datetime.now(timezone.utc)
                    CommandTopic.write_packet(command, scope=self.scope)
                    self.interface.write_raw(msg_hash[b"raw"])
                    return "SUCCESS"
                else:
                    return f"Interface not connected: {self.interface.name}"
            if msg_hash.get(b"log_stream"):
                if msg_hash[b"log_stream"].decode() == "true":
                    self.logger.info(f"{self.interface.name}: Enable stream logging")
                    self.interface.start_raw_logging()
                else:
                    self.logger.info(f"{self.interface.name}: Disable stream logging")
                    self.interface.stop_raw_logging()
                return "SUCCESS"
            if msg_hash.get(b"interface_cmd"):
                params = json.loads(msg_hash[b"interface_cmd"])
                try:
                    str_params = " ".join([str(i) for i in params["cmd_params"]])
                    self.logger.info(f"{self.interface.name}: interface_cmd: {params['cmd_name']} {str_params}")
                    self.interface.interface_cmd(params["cmd_name"], *params["cmd_params"])
                    InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
                except RuntimeError as error:
                    self.logger.error(f"{self.interface.name}: interface_cmd: {repr(error)}")
                    return repr(error)
                return "SUCCESS"
            if msg_hash.get(b"protocol_cmd"):
                params = json.loads(msg_hash[b"protocol_cmd"])
                try:
                    str_params = " ".join([str(i) for i in params["cmd_params"]])
                    self.logger.info(
                        f"{self.interface.name}: protocol_cmd: {params['cmd_name']} {str_params} read_write: {params['read_write']} index: {params['index']}"
                    )
                    self.interface.protocol_cmd(
                        params["cmd_name"],
                        *params["cmd_params"],
                        read_write=params["read_write"],
                        index=params["index"],
                    )
                    InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
                except RuntimeError as error:
                    self.logger.error(f"{self.interface.name}: protocol_cmd:{repr(error)}")
                    return repr(error)
                return "SUCCESS"
            if msg_hash.get(b"inject_tlm"):
                handle_inject_tlm(msg_hash[b"inject_tlm"], self.scope)
                return "SUCCESS"
            if msg_hash.get(b"release_critical"):
                model = CriticalCmdModel.get_model(name=msg_hash[b"release_critical"].decode(), scope=self.scope)
                if model is not None:
                    msg_hash = model.cmd_hash
                    release_critical = True
                else:
                    return f"Critical command {msg_hash[b'release_critical'].decode()} not found"

        target_name = msg_hash[b"target_name"].decode()
        cmd_name = msg_hash[b"cmd_name"].decode()
        manual = ConfigParser.handle_true_false(msg_hash[b"manual"].decode())
        cmd_params = None
        cmd_buffer = None
        hazardous_check = None
        if msg_hash[b"cmd_params"] is not None:
            cmd_params = json.loads(msg_hash[b"cmd_params"], cls=JsonDecoder)
            range_check = ConfigParser.handle_true_false(msg_hash[b"range_check"].decode())
            raw = ConfigParser.handle_true_false(msg_hash[b"raw"].decode())
            hazardous_check = ConfigParser.handle_true_false(msg_hash[b"hazardous_check"].decode())
        elif msg_hash[b"cmd_buffer"] is not None:
            cmd_buffer = msg_hash[b"cmd_buffer"]

        try:
            try:
                if cmd_params is not None:
                    command = System.commands.build_cmd(target_name, cmd_name, cmd_params, range_check, raw)
                elif cmd_buffer is not None:
                    if target_name:
                        command = System.commands.identify(cmd_buffer, [target_name])
                    else:
                        command = System.commands.identify(cmd_buffer, self.interface.cmd_target_names)
                    if not command:
                        command = System.commands.packet("UNKNOWN", "UNKNOWN").clone()
                        command.buffer = cmd_buffer
                else:
                    raise RuntimeError(f"Invalid command received:\n{msg_hash}")
                orig_command = System.commands.packet(command.target_name, command.packet_name)
                orig_command.received_count = TargetModel.increment_command_count(command.target_name, command.packet_name, 1, scope=self.scope)
                command.received_count = orig_command.received_count
                command.received_time = datetime.now(timezone.utc)
            except Exception as error:
                self.logger.error(f"{self.interface.name}: {msg_hash}")
                self.logger.error(f"{self.interface.name}: {repr(error)}")
                return repr(error)

            command.extra = command.extra or {}
            command.extra["cmd_string"] = msg_hash[b"cmd_string"].decode()
            command.extra["username"] = msg_hash[b"username"].decode()
            hazardous, hazardous_description = System.commands.cmd_pkt_hazardous(command)

            if hazardous_check:
                # Return back the error, description, and the formatted command
                # This allows the error handler to simply re-send the command
                if hazardous:
                    if not release_critical:
                        return f"HazardousError\n{hazardous_description}\n{command.extra['cmd_string']}"

            if self.critical_commanding is not None and self.critical_commanding != "OFF" and not release_critical:
                restricted = command.restricted
                if hazardous or restricted or (self.critical_commanding == "ALL" and manual):
                    type = None
                    if hazardous:
                        type = "HAZARDOUS"
                    elif restricted:
                        type = "RESTRICTED"
                    else:
                        type = "NORMAL"
                    model = CriticalCmdModel(
                        name=str(uuid.uuid1()),
                        type=type,
                        interface_name=self.interface.name,
                        username=msg_hash[b"username"].decode(),
                        cmd_hash=msg_hash,
                        scope=self.scope,
                    )
                    model.create()
                    self.logger.info(
                        f"Critical Cmd Pending: {msg_hash[b'cmd_string'].decode()}",
                        user=msg_hash[b"username"].decode(),
                        scope=self.scope,
                    )
                    return f"CriticalCmdError\n{model.name}"

            validate = ConfigParser.handle_true_false(msg_hash[b"validate"].decode())
            try:
                if self.interface.connected():
                    result = True
                    reason = None
                    if command.validator and validate:
                        try:
                            result, reason = command.validator.pre_check(command)
                        except Exception as error:
                            result = False
                            reason = repr(error)
                        if not result:
                            message = f"pre_check returned false for {command.extra['cmd_string']} due to {reason}"
                            raise WriteRejectError(message)

                    self.count += 1
                    if self.metric is not None:
                        self.metric.set(name="interface_cmd_total", value=self.count, type="counter")

                    log_message = ConfigParser.handle_true_false(msg_hash[b"log_message"].decode())
                    if log_message:
                        self.logger.info(
                            msg_hash[b"cmd_string"].decode(), user=msg_hash[b"username"].decode(), scope=self.scope
                        )

                    self.interface.write(command)

                    if command.validator and validate:
                        try:
                            result, reason = command.validator.post_check(command)
                        except Exception as error:
                            result = False
                            reason = repr(error)
                        command.extra["cmd_success"] = result
                        if reason:
                            command.extra["cmd_reason"] = reason

                    CommandDecomTopic.write_packet(command, scope=self.scope)
                    CommandTopic.write_packet(command, scope=self.scope)
                    InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)

                    if not result:
                        message = f"post_check returned false for {command.extra['cmd_string']} due to {reason}"
                        raise WriteRejectError(message)

                    return "SUCCESS"
                else:
                    return f"Interface not connected: {self.interface.name}"
            except WriteRejectError as error:
                return repr(error)
        except RuntimeError as error:
            self.logger.error(f"{self.interface.name}: {repr(error)}")
            return repr(error)


class RouterTlmHandlerThread:
    def __init__(self, router, tlm, logger=None, metric=None, scope=None):
        self.router = router
        self.tlm = tlm
        self.scope = scope
        self.logger = logger
        if not self.logger:
            self.logger = Logger
        self.metric = metric
        self.count = 0
        self.directive_count = 0
        if self.metric is not None:
            self.metric.set(
                name="router_directive_total",
                value=self.directive_count,
                type="counter",
            )
        if self.metric is not None:
            self.metric.set(name="router_tlm_total", value=self.count, type="counter")

    def start(self):
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.thread.start()
        ThreadManager.instance().register(self.thread, stop_object=self)
        return self.thread

    def stop(self):
        kill_thread(self, self.thread)

    def graceful_kill(self):
        RouterTopic.shutdown(self.router, scope=self.scope)
        time.sleep(0.001)  # Allow other threads to run

    def run(self):
        for topic, msg_id, msg_hash, redis in RouterTopic.receive_telemetry(self.router, scope=self.scope):
            msgid_seconds_from_epoch = int(msg_id.split("-")[0]) / 1000.0
            delta = time.time() - msgid_seconds_from_epoch
            if self.metric is not None:
                self.metric.set(
                    name="router_topic_delta_seconds",
                    value=delta,
                    type="gauge",
                    unit="seconds",
                    help="Delta time between data written to stream and router tlm start",
                )

            # Check for commands to the router itself
            if "CMD}ROUTER" in topic:
                self.directive_count += 1
                if self.metric is not None:
                    self.metric.set(
                        name="router_directive_total",
                        value=self.directive_count,
                        type="counter",
                    )

                if msg_hash.get(b"shutdown"):
                    self.logger.info(f"{self.router.name}: Shutdown requested")
                    RouterTopic.clear_topics(RouterTopic.topics(self.router, scope=self.scope))
                    return
                if msg_hash.get(b"connect"):
                    self.logger.info(f"{self.router.name}: Connect requested")
                    params = []
                    if msg_hash.get(b"params"):
                        params = json.loads(msg_hash[b"params"])
                    self.router = self.tlm.attempting(*params)
                if msg_hash.get(b"disconnect"):
                    self.logger.info(f"{self.router.name}: Disconnect requested")
                    self.tlm.disconnect(False)
                if msg_hash.get(b"log_stream"):
                    if msg_hash[b"log_stream"].decode() == "true":
                        self.logger.info(f"{self.router.name}: Enable stream logging")
                        self.router.start_raw_logging
                    else:
                        self.logger.info(f"{self.router.name}: Disable stream logging")
                        self.router.stop_raw_logging
                if msg_hash.get(b"router_cmd"):
                    params = json.loads(msg_hash[b"router_cmd"])
                    try:
                        self.logger.info(
                            f"{self.router.name}: router_cmd: {params['cmd_name']} {' '.join(params['cmd_params'])}"
                        )
                        self.router.interface_cmd(params["cmd_name"], *params["cmd_params"])
                    except RuntimeError as error:
                        self.logger.error(f"{self.router.name}: router_cmd: {repr(error)}")
                        return error.message
                    return "SUCCESS"
                if msg_hash.get(b"protocol_cmd"):
                    params = json.loads(msg_hash[b"protocol_cmd"])
                    try:
                        self.logger.info(
                            f"{self.router.name}: protocol_cmd: {params['cmd_name']} {' '.join(params['cmd_params'])} read_write: {params['read_write']} index: {params['index']}"
                        )
                        self.router.protocol_cmd(
                            params["cmd_name"],
                            *params["cmd_params"],
                            read_write=params["read_write"],
                            index=params["index"],
                        )
                    except RuntimeError as error:
                        self.logger.error(f"{self.router.name}: protoco_cmd: {repr(error)}")
                        return error.message
                    return "SUCCESS"
                return "SUCCESS"

            if self.router.connected():
                self.count += 1
                if self.metric is not None:
                    self.metric.set(name="router_tlm_total", value=self.count, type="counter")

                target_name = msg_hash[b"target_name"].decode()
                packet_name = msg_hash[b"packet_name"].decode()

                packet = System.telemetry.packet(target_name, packet_name)
                packet.stored = ConfigParser.handle_true_false(msg_hash[b"stored"].decode())
                packet.received_time = from_nsec_from_epoch(int(msg_hash[b"time"]))
                packet.received_count = int(msg_hash[b"received_count"])
                packet.buffer = msg_hash[b"buffer"]

                try:
                    self.router.write(packet)
                    RouterStatusModel.set(self.router.as_json(), queued=True, scope=self.scope)
                    return "SUCCESS"
                except RuntimeError as error:
                    self.logger.error(f"{self.router.name}: {repr(error)}")
                    return repr(error)


class InterfaceMicroservice(Microservice):
    UNKNOWN_BYTES_TO_PRINT = 16
    DISCONNECT_WAIT_TIME = 1

    def __init__(self, name):
        self.mutex = threading.Lock()
        self.interface = None
        self.interface_or_router = None
        self.interface_thread_sleeper = Sleeper()
        self.cancel_thread = False
        self.connection_failed_messages = []
        self.connection_lost_messages = []
        self.handler_thread = None

        super().__init__(name)
        self.interface_or_router = self.__class__.__name__.split("Microservice")[0].upper()
        if self.interface_or_router == "INTERFACE":
            self.metric.set(name="interface_tlm_total", value=self.count, type="counter")
        else:
            self.metric.set(name="router_cmd_total", value=self.count, type="counter")

        self.scope = name.split("__")[0]
        interface_name = name.split("__")[2]
        if self.interface_or_router == "INTERFACE":
            self.interface = InterfaceModel.get_model(name=interface_name, scope=self.scope).build()
        else:
            self.interface = RouterModel.get_model(name=interface_name, scope=self.scope).build()
        self.interface.name = interface_name
        # Map the interface to the interface's targets
        for target_name in self.interface.target_names:
            target = System.targets[target_name]
            target.interface = self.interface
        for target_name in self.interface.tlm_target_names:
            # Initialize the target's packet counters based on the Topic stream
            # Prevents packet count resetting to 0 when interface restarts
            try:
                for packet_name, packet in System.telemetry.packets(target_name).items():
                    topic = f"{self.scope}__TELEMETRY__{{target_name}}__{packet_name}"
                    msg_id, msg_hash = Topic.get_newest_message(topic)
                    if msg_id:
                        packet.received_count = int(msg_hash[b"received_count"])
                    else:
                        packet.received_count = 0
            except RuntimeError:
                pass  # Handle targets without telemetry
        if self.interface.connect_on_startup:
            self.interface.state = "ATTEMPTING"
        else:
            self.interface.state = "DISCONNECTED"
        if self.interface_or_router == "INTERFACE":
            InterfaceStatusModel.set(self.interface.as_json(), scope=self.scope)
        else:
            RouterStatusModel.set(self.interface.as_json(), scope=self.scope)

        self.queued = False
        self.sync_packet_count_data = {}
        self.sync_packet_count_time = None
        self.sync_packet_count_delay_seconds = 1.0 # Sync packet counts every second
        for option_name, option_values in self.interface.options.items():
            if option_name.upper() == "OPTIMIZE_THROUGHPUT":
                self.queued = True
                update_interval = float(option_values[0])
                EphemeralStoreQueued.instance().set_update_interval(update_interval)
                StoreQueued.instance().set_update_interval(update_interval)
            if option_name.upper() == 'SYNC_PACKET_COUNT_DELAY_SECONDS':
                self.sync_packet_count_delay_seconds = float(option_values[0])

        if self.interface_or_router == "INTERFACE":
            self.handler_thread = InterfaceCmdHandlerThread(
                self.interface,
                self,
                logger=self.logger,
                metric=self.metric,
                scope=self.scope,
            )
        else:
            self.handler_thread = RouterTlmHandlerThread(
                self.interface,
                self,
                logger=self.logger,
                metric=self.metric,
                scope=self.scope,
            )
        self.handler_thread.start()

    # Called to connect the interface/router. It takes optional parameters to
    # rebuilt the interface/router. Once we set the state to 'ATTEMPTING' the
    # run method handles the actual connection.
    def attempting(self, *params):
        try:
            if len(params) != 0:
                self.interface.disconnect()
                # Build New Interface, this can fail if passed bad parameters
                new_interface = self.interface.__class__(*params)
                self.interface.copy_to(new_interface)

                # Replace interface for targets
                for target_name in self.interface.target_names:
                    target = System.targets[target_name]
                    target.interface = new_interface
                self.interface = new_interface

            self.interface.state = "ATTEMPTING"
            if self.interface_or_router == "INTERFACE":
                InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
            else:
                RouterStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
            return self.interface  # Return the interface/router since we may have recreated it
        # Need to rescue Exception so we cover LoadError
        except RuntimeError as error:
            self.logger.error(f"Attempting connection #{self.interface.connection_string} failed due to {repr(error)}")
            # if SignalException === error:
            #   self.logger.info(f"{self.interface.name}: Closing from signal")
            #   self.cancel_thread = True
            return self.interface  # Return the original interface/router in match of error:

    def run(self):
        try:
            if self.interface.read_allowed:
                self.logger.info(f"{self.interface.name}: Starting packet reading")
            else:
                self.logger.info(f"{self.interface.name}: Starting connection maintenance")
            while True:
                if self.cancel_thread:
                    break

                match self.interface.state:
                    case "DISCONNECTED":
                        # Just wait to see if we should connect later
                        self.interface_thread_sleeper.sleep(InterfaceMicroservice.DISCONNECT_WAIT_TIME)
                    case "ATTEMPTING":
                        try:
                            with self.mutex:
                                # We need to make sure connect is not called after stop() has been called
                                if not self.cancel_thread:
                                    self.connect()
                        except (RuntimeError, OSError) as error:
                            self.handle_connection_failed(self.interface.connection_string(), error)
                            if self.cancel_thread:
                                break
                    case "CONNECTED":
                        if self.interface.read_allowed:
                            try:
                                packet = self.interface.read()
                                if packet is not None:
                                    self.handle_packet(packet)
                                    self.count += 1
                                    if self.interface_or_router == "INTERFACE":
                                        self.metric.set(
                                            name="interface_tlm_total",
                                            value=self.count,
                                            type="counter",
                                        )
                                    else:
                                        self.metric.set(
                                            name="router_cmd_total",
                                            value=self.count,
                                            type="counter",
                                        )
                                else:
                                    self.logger.info(
                                        f"{self.interface.name}: Internal disconnect requested (returned None)"
                                    )
                                    self.handle_connection_lost()
                                    if self.cancel_thread:
                                        break
                            except RuntimeError as error:
                                self.handle_connection_lost(error)
                                if self.cancel_thread:
                                    break
                        else:
                            self.interface_thread_sleeper.sleep(1)
                            if not self.interface.connected():
                                self.handle_connection_lost()
        except RuntimeError as error:
            if not isinstance(error, SystemExit):  # or signal exception
                self.logger.error(f"{self.interface.name}: Packet reading thread died: {repr(error)}")
                # handle_fatal_exception(error)
            # Try to do clean disconnect because we're going down
            self.disconnect(False)
        if self.interface_or_router == "INTERFACE":
            InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
        else:
            RouterStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
        self.logger.info(f"{self.interface.name}: Stopped packet reading")

    def handle_packet(self, packet):
        InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
        if packet.received_time is None:
            packet.received_time = datetime.now(timezone.utc)

        if packet.stored:
            # Stored telemetry does not update the current value table
            identified_packet = System.telemetry.identify_and_define_packet(packet, self.interface.tlm_target_names)
        else:
            # Identify and update packet
            if packet.identified():
                try:
                    # Preidentifed packet - place it into the current value table
                    identified_packet = System.telemetry.update(packet.target_name, packet.packet_name, packet.buffer)
                except RuntimeError:
                    # Packet identified but we don't know about it
                    # Clear packet_name and target_name and try to identify
                    self.logger.warn(
                        f"{self.interface.name}: Received unknown identified telemetry: {packet.target_name} {packet.packet_name}"
                    )
                    packet.target_name = None
                    packet.packet_name = None
                    identified_packet = System.telemetry.identify_and_set_buffer(
                        packet.buffer, self.interface.tlm_target_names
                    )
            else:
                # Packet needs to be identified
                identified_packet = System.telemetry.identify_and_set_buffer(
                    packet.buffer, self.interface.tlm_target_names
                )

        if identified_packet:
            identified_packet.received_time = packet.received_time
            identified_packet.stored = packet.stored
            identified_packet.extra = packet.extra
            packet = identified_packet
        else:
            unknown_packet = System.telemetry.update("UNKNOWN", "UNKNOWN", packet.buffer)
            unknown_packet.received_time = packet.received_time
            unknown_packet.stored = packet.stored
            unknown_packet.extra = packet.extra
            packet = unknown_packet
            json_hash = CvtModel.build_json_from_packet(packet)
            CvtModel.set(
                json_hash,
                packet.target_name,
                packet.packet_name,
                queued=self.queued,
                scope=self.scope,
            )
            num_bytes_to_print = min(InterfaceMicroservice.UNKNOWN_BYTES_TO_PRINT, len(packet.buffer))
            data = packet.buffer_no_copy()[0:(num_bytes_to_print)]
            prefix = "".join([format(x, "02x") for x in data])
            self.logger.warn(
                f"{self.interface.name} {packet.target_name} packet length: {len(packet.buffer)} starting with: {prefix}"
            )

        # Write to stream
        self.sync_tlm_packet_counts(packet)
        TelemetryTopic.write_packet(packet, queued=self.queued, scope=self.scope)

    def handle_connection_failed(self, connection, connect_error):
        self.error = connect_error
        self.logger.error(f"{self.interface.name}: Connection {connection} failed due to {repr(connect_error)}")
        # match connect_error:
        #   case OSError:
        #     self.logger.info(f"{self.interface.name}: Closing from signal")
        #     self.cancel_thread = True
        # case Errno='ECONNREFUSED', Errno='ECONNRESET', Errno='ETIMEDOUT', Errno='ENOTSOCK', Errno='EHOSTUNREACH', IOError:
        #   # Do not write an exception file for these extremely common cases
        # else _:
        if connect_error is RuntimeError and ("canceled" in connect_error.message or "timeout" in repr(connect_error)):
            pass  # Do not write an exception file for these extremely common cases
        else:
            self.logger.error(f"{self.interface.name}: {str(connect_error)}")
            if str(connect_error) not in self.connection_failed_messages:
                # OpenC3.write_exception_file(connect_error)
                self.connection_failed_messages.append(str(connect_error))
        self.disconnect()  # Ensure we do a clean disconnect

    def handle_connection_lost(self, error=None, reconnect=True):
        if error:
            self.error = error
            self.logger.info(f"{self.interface.name}: Connection Lost: {repr(error)}")
            # match err:
            #   case SignalException:
            #     self.logger.info(f"{self.interface.name}: Closing from signal")
            #     self.cancel_thread = True
            #   # case Errno='ECONNABORTED', Errno='ECONNRESET', Errno='ETIMEDOUT', Errno='EBADF', Errno='ENOTSOCK', IOError:
            #     # Do not write an exception file for these extremely common cases
            #   else _:
            self.logger.error(f"{self.interface.name}: {str(error)}")
            if str(error) not in self.connection_lost_messages:
                # OpenC3.write_exception_file(err)
                self.connection_lost_messages.append(str(error))
        else:
            self.logger.info(f"{self.interface.name}: Connection Lost")
        self.disconnect(reconnect)  # Ensure we do a clean disconnect

    def connect(self):
        self.logger.info(f"{self.interface.name}: Connect {self.interface.connection_string()}")

        try:
            self.interface.connect()
            self.interface.post_connect()
        except RuntimeError as error:
            try:
                self.interface.disconnect()  # Ensure disconnect is called at least once on a partial connect
            except RuntimeError:
                pass  # We want to report any connect errors, not disconnect in this case
            raise error

        self.interface.state = "CONNECTED"
        if self.interface_or_router == "INTERFACE":
            InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
        else:
            RouterStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
        self.logger.info(f"{self.interface.name}: Connection Success")

    def disconnect(self, allow_reconnect=True):
        if self.interface.state == "DISCONNECTED" and self.interface.connected() is False:
            return

        # Synchronize the calls to @interface.disconnect since it takes an unknown
        # amount of time. If two calls to disconnect stack up, the if statement
        # should avoid multiple calls to disconnect.
        with self.mutex:
            try:
                if self.interface.connected():
                    self.interface.disconnect()
            except RuntimeError as error:
                self.logger.error(f"Disconnect: {self.interface.name}: {repr(error)}")

        # If the interface is set to auto_reconnect then delay so the thread
        # can come back around and allow the interface a chance to reconnect.
        if allow_reconnect and self.interface.auto_reconnect and self.interface.state != "DISCONNECTED":
            self.attempting()
            if self.cancel_thread is not None:
                self.interface_thread_sleeper.sleep(self.interface.reconnect_delay)
        else:
            self.interface.state = "DISCONNECTED"
            if self.interface_or_router == "INTERFACE":
                InterfaceStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)
            else:
                RouterStatusModel.set(self.interface.as_json(), queued=True, scope=self.scope)

    # Disconnect from the interface and stop the thread
    def stop(self):
        name = self.name
        if self.interface:
            name = self.interface.name
        self.logger.info(f"{name}: stop requested")
        with self.mutex:
            # Need to make sure that @cancel_thread is set and the interface disconnected within
            # mutex to ensure that connect() is not called when we want to stop()
            self.cancel_thread = True
            if self.handler_thread:
                self.handler_thread.stop()
            if self.interface_thread_sleeper:
                self.interface_thread_sleeper.cancel()
            if self.interface:
                self.interface.disconnect()
                if self.interface_or_router == "INTERFACE":
                    valid_interface = InterfaceStatusModel.get_model(name=self.interface.name, scope=self.scope)
                else:
                    valid_interface = RouterStatusModel.get_model(name=self.interface.name, scope=self.scope)
                if valid_interface:
                    valid_interface.destroy()

    def shutdown(self, sig=None):
        if self.shutdown_complete:
            return  # Nothing more to do
        name = self.name
        if self.interface:
            name = self.interface.name
        self.logger.info(f"{name}: shutdown requested")
        self.stop()
        super().shutdown()

    def graceful_kill(self):
        pass  # Just to avoid warning

    def sync_tlm_packet_counts(self, packet):
        if self.sync_packet_count_delay_seconds <= 0:
            # Perfect but slow method
            packet.received_count = TargetModel.increment_telemetry_count(packet.target_name, packet.packet_name, 1, scope=self.scope)
        else:
            # Eventually consistent method
            # Only sync every period (default 1 second) to avoid hammering Redis
            # This is a trade off between speed and accuracy
            # The packet count is eventually consistent
            if self.sync_packet_count_data.get(packet.target_name) is None:
                self.sync_packet_count_data[packet.target_name] = {}
            if self.sync_packet_count_data[packet.target_name].get(packet.packet_name) is None:
                self.sync_packet_count_data[packet.target_name][packet.packet_name] = 0
            self.sync_packet_count_data[packet.target_name][packet.packet_name] += 1

            # Ensures counters change between syncs
            packet.received_count += 1

            # Check if we need to sync the packet counts
            if self.sync_packet_count_time is None or (time.time() - self.sync_packet_count_time) > self.sync_packet_count_delay_seconds:
                self.sync_packet_count_time = time.time()

                result = []
                inc_count = 0
                # Use pipeline to make this one transaction
                with Store.instance().redis_pool.get() as redis:
                    pipeline = redis.pipeline(transaction=False)
                    thread_id = threading.get_native_id()
                    Store.instance().redis_pool.pipelines[thread_id] = pipeline
                    try:
                        # Increment global counters for packets received
                        for target_name, packet_data in self.sync_packet_count_data.items():
                            for packet_name, count in packet_data.items():
                                TargetModel.increment_telemetry_count(target_name, packet_name, count, scope=self.scope)
                                inc_count += 1
                        self.sync_packet_count_data = {}

                        # Get all the packet counts with the global counters
                        for target_name in self.interface.tlm_target_names:
                            TargetModel.get_all_telemetry_counts(target_name, scope=self.scope)
                        TargetModel.get_all_telemetry_counts('UNKNOWN', scope=self.scope)

                        result = pipeline.execute()
                    finally:
                        Store.instance().redis_pool.pipelines[thread_id] = None
                for target_name in self.interface.tlm_target_names:
                    for packet_name, count in result[inc_count].items():
                        update_packet = System.telemetry.packet(target_name, packet_name.decode())
                        update_packet.received_count = int(count)
                    inc_count += 1
                for packet_name, count in result[inc_count].items():
                    update_packet = System.telemetry.packet('UNKNOWN', packet_name.decode())
                    update_packet.received_count = int(count)

if os.path.basename(__file__) == os.path.basename(sys.argv[0]):
    InterfaceMicroservice.class_run()
    ThreadManager.instance().shutdown()
    ThreadManager.instance().join()


# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


class FormatStringParser:
    # self.param parser [ConfigParser] Configuration parser
    # self.param item [Packet] The current item
    @classmethod
    def parse(cls, parser, item):
        parser = FormatStringParser(parser)
        parser.verify_parameters()
        parser.create_format_string(item)

    # self.param parser [ConfigParser] Configuration parser
    def __init__(self, parser):
        self.parser = parser

    def verify_parameters(self):
        self.usage = "FORMAT_STRING <PRINTF STYLE STRING>"
        self.parser.verify_num_parameters(1, 1, self.usage)

    # self.param item [PacketItem] The item the limits response should be added to
    def create_format_string(self, item):
        item.format_string = self.parser.parameters[0]
        # Only test the format string if there is not a read conversion because:
        # read conversion can return any type
        if item.read_conversion is None:
            self._test_format_string(item)

    def _test_format_string(self, item):
        try:
            match item.data_type:
                case "INT" | "UINT":
                    f"{item.format_string}" % 0
                case "FLOAT":
                    f"{item.format_string}" % 0.0
                case "STRING" | "BLOCK":
                    f"{item.format_string}" % "Hello"
        except TypeError as error:
            raise self.parser.error(
                f"Invalid FORMAT_STRING specified for type {item.data_type}: {self.parser.parameters[0]}",
                self.usage,
            ) from error



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from datetime import datetime, timezone
from openc3.conversions.conversion import Conversion


# Converts a unix format time: Epoch Jan 1 1970, seconds and microseconds
class UnixTimeConversion(Conversion):
    # Initializes the time item to grab from the packet
    #
    # @param seconds_item_name [String] The telemetry item in the packet which
    #   represents the number of seconds since the UNIX time epoch
    # @param microseconds_item_name [String] The telemetry item in the packet
    #   which represents microseconds
    def __init__(self, seconds_item_name, microseconds_item_name=None):
        super().__init__()
        self.seconds_item_name = seconds_item_name
        self.microseconds_item_name = microseconds_item_name
        self.converted_type = "RUBY_TIME"
        self.converted_bit_size = 0
        self.params = [seconds_item_name, microseconds_item_name]

    # @param (see Conversion#call)
    # @return [Float] Packet time in seconds since UNIX epoch
    def call(self, value, packet, buffer):
        time = packet.read(self.seconds_item_name, "RAW", buffer)
        if self.microseconds_item_name:
            time += packet.read(self.microseconds_item_name, "RAW", buffer) / 1000000.0
        return datetime.fromtimestamp(time, tz=timezone.utc)

    # @return [String] The name of the class followed by the time conversion
    def __str__(self):
        return f"UnixTimeConversion {self.seconds_item_name} {self.microseconds_item_name}"

    # @param (see Conversion#to_config)
    # @return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        return f"    {read_or_write}_CONVERSION {self.__class__.__name__} {self.seconds_item_name} {self.microseconds_item_name}\n"



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from websockets.sync.client import connect


class WebSocketClientStream:
    # @param url [String] The host to connect to
    # @param write_timeout [Float] Seconds to wait before aborting writes
    # @param read_timeout [Float|nil] Seconds to wait before aborting reads.
    #   Pass nil to block until the read is complete.
    # @param connect_timeout [Float|nil] Seconds to wait before aborting connect.
    #   Pass nil to block until the connection is complete.
    def __init__(self, url, write_timeout, read_timeout, connect_timeout=5.0):
        self.url = url
        self.recv_timeout = read_timeout
        self.connect_timeout = connect_timeout
        self.headers = {}

    def connect(self):
        self.connection = connect(
            self.url,
            subprotocols=["actioncable-v1-json"],
            open_timeout=self.connect_timeout,
            additional_headers=self.headers,
        )
        return True

    def read(self):
        while True:
            try:
                return self.connection.recv(self.recv_timeout)
            except TimeoutError:
                return None

    def write(self, data):
        self.connection.send(data)

    def connected(self):
        if self.connection:
            return True
        return False

    def disconnect(self):
        if self.connection:
            self.connection.close()



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from openc3.conversions.conversion import Conversion
from openc3.config.config_parser import ConfigParser
from openc3.accessors.binary_accessor import BinaryAccessor


class ProcessorConversion(Conversion):
    # @param processor_name [String] The name of the associated processor
    # @param result_name [String] The name of the associated result in the processor
    # @param converted_type [String or nil] The datatype of the result of the processor
    # @param converted_bit_size [Integer or nil] The bit size of the result of the processor
    # @param converted_array_size [Integer or nil] The total array bit size of the result of the processor
    def __init__(
        self,
        processor_name,
        result_name,
        converted_type=None,
        converted_bit_size=None,
        converted_array_size=None,
    ):
        super().__init__()
        self.processor_name = str(processor_name).upper()
        self.result_name = str(result_name).upper()
        self.params = [self.processor_name, self.result_name]
        if ConfigParser.handle_none(converted_type):
            self.converted_type = str(converted_type).upper()
            if self.converted_type not in BinaryAccessor.DATA_TYPES:
                raise TypeError(f"Unknown converted type: {converted_type}")
            self.params.append(self.converted_type)
        if ConfigParser.handle_none(converted_bit_size):
            self.converted_bit_size = int(converted_bit_size)
            self.params.append(self.converted_bit_size)
        if ConfigParser.handle_none(converted_array_size):
            self.converted_array_size = int(converted_array_size)
            self.params.append(self.converted_array_size)

    # @param (see Conversion#call)
    # @return [Varies] The result of the associated processor
    def call(self, value, packet, buffer):
        result = packet.processors[self.processor_name].results[self.result_name]
        if result is not None:
            return result
        else:
            return 0

    # @return [String] The type of processor
    def __str__(self):
        return f"ProcessorConversion {self.processor_name} {self.result_name}"

    # @param (see Conversion#to_config)
    # @return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        config = f"    {read_or_write}_CONVERSION {self.__class__.__name__} {self.processor_name} {self.result_name}"
        if self.converted_type is not None:
            config += f" {self.converted_type}"
        if self.converted_bit_size is not None:
            config += f" {self.converted_bit_size}"
        if self.converted_array_size is not None:
            config += f" {self.converted_array_size}"
        config += "\n"
        return config



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import queue
from threading import Thread
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
from openc3.interfaces.interface import Interface
from openc3.accessors.http_accessor import HttpAccessor
from openc3.packets.packet import Packet
from openc3.system.system import System
from openc3.utilities.logger import Logger


class Handler(BaseHTTPRequestHandler):
    def do_POST(self):
        self.handle_request()

    def do_GET(self):
        self.handle_request()

    def handle_request(self):
        base = self.path.split("?")[0]
        if self.server.lookup.get(base):
            packets = self.server.lookup[base]
            status = 200

            for packet in packets:
                # Build the Response
                try:
                    status = int(packet.read("HTTP_STATUS"))
                except Exception:
                    # No HTTP_STATUS - Leave at default
                    pass

                self.send_response(status)

                # http_accessor stores all the pseudo-derived HTTP configuration in extra
                if packet.extra:
                    headers = packet.extra.get("HTTP_HEADERS")
                    if headers:
                        for key, value in headers:
                            self.send_header(key, value)
                self.end_headers()

                self.wfile.write(packet.buffer)

                # Save the Request
                packet_name = None
                try:
                    packet_name = packet.read("HTTP_PACKET")
                except Exception:
                    # No packet name means don't save the request as telemetry
                    pass
                if packet_name:
                    data = b""
                    if self.headers.get("content-length") and self.rfile:
                        length = int(self.headers.get("content-length"))
                        data = self.rfile.read(length)
                    extra = {}
                    extra["HTTP_REQUEST_TARGET_NAME"] = packet.target_name
                    extra["HTTP_REQUEST_PACKET_NAME"] = packet_name

                    if self.headers:
                        extra["HTTP_HEADERS"] = {}
                        for key, value in self.headers.items():
                            extra["HTTP_HEADERS"][key.lower()] = value

                    queries = {}
                    if "?" in self.path:
                        query_string = self.path.split("?", 1)[1]
                        for query in query_string.split("&"):
                            key, value = query.split("=", 1)
                            queries[key] = value

                    if len(queries) > 0:
                        extra["HTTP_QUERIES"] = {}
                        for key, value in queries.items():
                            extra["HTTP_QUERIES"][key] = value

                    self.server.request_queue.put((data, extra))


class HttpServerInterface(Interface):
    # @param port [Integer] HTTP port
    def __init__(self, port=80):
        super().__init__()
        self.listen_address = "0.0.0.0"  # Default to ANY
        self.port = int(port)
        self.server = None

    def build_path_lookup(self):
        self.lookup = {}
        for target_name in self.target_names:
            for packet_name, packet in System.commands.packets(target_name).items():
                packet.restore_defaults()
                path = None
                try:
                    path = packet.read("HTTP_PATH")
                except Exception as error:
                    # No HTTP_PATH is an error
                    Logger.error(
                        f"HttpServerInterface Packet {target_name} {packet_name} unable to read HTTP_PATH\n{repr(error)}"
                    )
                if path:
                    self.lookup[path] = self.lookup.get(path) or []
                    self.lookup[path].append(packet)

    # Supported Options
    # LISTEN_ADDRESS - Ip address of the interface to accept connections on
    # (see Interface#set_option)
    def set_option(self, option_name, option_values):
        super().set_option(option_name, option_values)
        if option_name.upper() == "LISTEN_ADDRESS":
            self.listen_address = option_values[0]

    def connection_string(self):
        return f"listening on {self.listen_address}:{self.port}"

    # Connects the interface to its target(s)
    def connect(self):
        # Can't build the lookup until after init because the target_names are not set
        self.build_path_lookup()
        self.server = ThreadingHTTPServer((self.listen_address, self.port), Handler)
        self.server.request_queue = queue.Queue()
        self.server.lookup = self.lookup
        self.server_thread = Thread(target=self.server.serve_forever)
        self.server_thread.start()
        super().connect()

    def connected(self):
        if self.server:
            return True
        else:
            return False

    # Disconnects the interface from its target(s)
    def disconnect(self):
        if self.server:
            self.server.shutdown()
            self.server_thread.join()
        self.server = None
        super().disconnect()

    def convert_packet_to_data(self, packet):
        raise RuntimeError("Commands cannot be sent to HttpServerInterface")

    def write_interface(self, data, extra=None):
        raise RuntimeError("Commands cannot be sent to HttpServerInterface")

    def read_interface(self):
        data, extra = self.server.request_queue.get(block=True)
        if data is None:
            return data, extra
        self.read_interface_base(data, extra)
        return data, extra

    # Called to convert the read data into a OpenC3 Packet object
    #
    # @param data [String] Raw packet data
    # @return [Packet] OpenC3 Packet with buffer filled with data
    def convert_data_to_packet(self, data, extra=None):
        packet = Packet(None, None, "BIG_ENDIAN", None, data)
        packet.accessor = HttpAccessor(packet)
        if extra:
            # Identify the response
            request_target_name = extra["HTTP_REQUEST_TARGET_NAME"]
            request_packet_name = extra["HTTP_REQUEST_PACKET_NAME"]
            if request_target_name and request_packet_name:
                packet.target_name = str(request_target_name).upper()
                packet.packet_name = str(request_packet_name).upper()
            extra.pop("HTTP_REQUEST_TARGET_NAME", None)
            extra.pop("HTTP_REQUEST_PACKET_NAME", None)
            packet.extra = extra

        return packet



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


# Class that implements the following methods= read, write(data),
# connect, connected? and disconnect. Streams are simply data sources which
# {Protocol} classes read and write to. This separation of concerns
# allows Streams to simply focus on getting and sending raw data while the
# higher level processing occurs in {Protocol}.
class Stream:
    # Connects the stream
    def connect(self):
        raise RuntimeError("connect not defined by Stream")

    # Whether the stream is connected
    def connected(self):
        raise RuntimeError("connected not defined by Stream")

    # Disconnects the stream
    # Note that streams are not designed to be reconnected and must be recreated
    def disconnect(self):
        raise RuntimeError("disconnect not defined by Stream")

    # Expected to return any amount of data on success, or a blank string on
    # closed/EOF, and may raise Timeout='E'rror, or other errors
    def read(self):
        raise RuntimeError("read not defined by Stream")

    # Expected to always return immediately with data if available or an empty string
    # Should not raise errors
    def read_nonblock(self):
        raise RuntimeError("read_nonblock not defined by Stream")

    # Expected to write complete set of data.  May raise TimeoutError
    # or other errors.
    #
    # self.param data [String] Binary data to write to the stream
    def write(self, data):
        raise RuntimeError("write not defined by Stream")



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import json
from openc3.environment import OPENC3_LOCAL_MODE_PATH


class LocalMode:
    LOCAL_MODE_PATH = OPENC3_LOCAL_MODE_PATH or "/plugins"
    # When updating update local_mode.rb, PluginsTab.vue, plugins.spec.ts
    DEFAULT_PLUGINS = [
        "openc3-cosmos-tool-admin",
        "openc3-cosmos-tool-bucketexplorer",
        "openc3-cosmos-tool-cmdsender",
        "openc3-cosmos-tool-cmdhistory",
        "openc3-cosmos-tool-cmdtlmserver",
        "openc3-cosmos-tool-dataextractor",
        "openc3-cosmos-tool-dataviewer",
        "openc3-cosmos-tool-docs",
        "openc3-cosmos-tool-handbooks",
        "openc3-cosmos-tool-iframe",
        "openc3-cosmos-tool-limitsmonitor",
        "openc3-cosmos-tool-packetviewer",
        "openc3-cosmos-tool-scriptrunner",
        "openc3-cosmos-tool-tablemanager",
        "openc3-cosmos-tool-tlmgrapher",
        "openc3-cosmos-tool-tlmviewer",
        "openc3-cosmos-enterprise-tool-admin",
        "openc3-cosmos-tool-autonomic",
        "openc3-cosmos-tool-calendar",
        "openc3-cosmos-tool-grafana",
        "openc3-enterprise-tool-base",
        "openc3-tool-base",
    ]

    @classmethod
    def put_target_file(cls, path, io_or_string, scope):
        full_folder_path = f"{cls.LOCAL_MODE_PATH}/{path}"
        if not os.path.normpath(full_folder_path).startswith(cls.LOCAL_MODE_PATH):
            return
        os.makedirs(os.path.dirname(full_folder_path), exist_ok=True)
        flags = "w"
        if isinstance(io_or_string, (bytes, bytearray)):
            flags += "b"
        with open(full_folder_path, flags) as file:
            if hasattr(io_or_string, "read"):
                data = io_or_string.read()
            else:  # str or bytes
                data = io_or_string
            file.write(data)

    @classmethod
    def open_local_file(cls, path, scope):
        try:
            full_path = f"{cls.LOCAL_MODE_PATH}/{scope}/targets_modified/{path}"
            if os.path.normpath(full_path).startswith(cls.LOCAL_MODE_PATH):
                return open(full_path, "rb")
            return None
        except OSError:
            return None

    @classmethod
    def save_tool_config(cls, scope, tool, name, data):
        json_data = json.loads(data)
        config_path = f"{cls.LOCAL_MODE_PATH}/{scope}/tool_config/{tool}/{name}.json"
        if os.path.normpath(config_path).startswith(cls.LOCAL_MODE_PATH):
            os.makedirs(os.path.dirname(config_path), exist_ok=True)
            with open(config_path, "w") as file:
                file.write(json.dumps(json_data, indent=2))

    @classmethod
    def delete_tool_config(cls, scope, tool, name):
        config_path = f"{cls.LOCAL_MODE_PATH}/{scope}/tool_config/{tool}/{name}.json"
        if os.path.normpath(config_path).startswith(cls.LOCAL_MODE_PATH):
            os.remove(config_path)

    @classmethod
    def save_setting(cls, scope, name, data):
        config_path = f"{cls.LOCAL_MODE_PATH}/{scope}/settings/{name}.json"
        if os.path.normpath(config_path).startswith(cls.LOCAL_MODE_PATH):
            os.makedirs(os.path.dirname(config_path), exist_ok=True)
            # Anything can be stored as a setting so write it out directly
            with open(config_path, "w") as file:
                file.write(str(data))



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import selectors


# Allows for a breakable sleep implementation using the self-pipe trick
# See http://www.sitepoint.com/the-self-pipe-trick-explained/
class Sleeper:
    def __init__(self):
        self.pipe_reader, self.pipe_writer = os.pipe()
        self.selector = selectors.DefaultSelector()
        self.selector.register(self.pipe_reader, selectors.EVENT_READ)
        self.canceled = False

    # Breakable version of sleep
    # @param seconds Number of seconds to sleep
    # @return true if the sleep was broken by someone calling cancel
    #   otherwise returns false
    def sleep(self, seconds):
        list = self.selector.select(timeout=seconds)
        if list and list[0]:
            try:
                os.close(self.pipe_reader)
            except Exception:
                pass
            return True
        else:
            return False

    # Break sleeping - Once canceled a sleeper cannot be used again
    def cancel(self):
        if not self.canceled:
            self.canceled = True
            os.write(self.pipe_writer, bytes(".", encoding="ascii"))
            os.close(self.pipe_writer)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.environment import OPENC3_SCOPE
from openc3.models.model import Model
from openc3.utilities.store import Store, EphemeralStore


class MetricModel(Model):
    PRIMARY_KEY = "__openc3__metric"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str):
        return super().get(f"{scope}{MetricModel.PRIMARY_KEY}", name=name)

    @classmethod
    def names(cls, scope: str):
        return super().names(f"{scope}{MetricModel.PRIMARY_KEY}")

    @classmethod
    def all(cls, scope: str):
        return super().all(f"{scope}{MetricModel.PRIMARY_KEY}")

    # Sets (updates) the redis hash of this model
    @classmethod
    def set(cls, json: dict, scope: str = OPENC3_SCOPE, queued: bool = True):
        json["scope"] = scope
        cls(**json).create(force=True, queued=queued)

    @classmethod
    def destroy(cls, scope: str, name: str):
        EphemeralStore.hdel(f"{scope}{MetricModel.PRIMARY_KEY}", name)

    def __init__(self, name: str, values: dict = None, scope: str = OPENC3_SCOPE):
        values = {} if values is None else values
        super().__init__(f"{scope}{MetricModel.PRIMARY_KEY}", name=name, scope=scope)
        self.values = values

    def as_json(self):
        return {"name": self.name, "updated_at": self.updated_at, "values": self.values}

    @classmethod
    def redis_extract_p50_and_p99_seconds(cls, value):
        if value:
            split_value = str(value).split(",")
            p50 = float(split_value[0].split("=")[-1]) / 1_000_000
            p99 = float(split_value[-1].split("=")[-1]) / 1_000_000
            return p50, p99
        else:
            return 0.0, 0.0

    @classmethod
    def redis_metrics(cls):
        result = {}

        metrics = Store.info("all")
        result["redis_connected_clients_total"] = metrics["connected_clients"]
        result["redis_used_memory_rss_total"] = metrics["used_memory_rss"]
        result["redis_commands_processed_total"] = metrics["total_commands_processed"]
        result["redis_iops"] = metrics["instantaneous_ops_per_sec"]
        result["redis_instantaneous_input_kbps"] = metrics["instantaneous_input_kbps"]
        result["redis_instantaneous_output_kbps"] = metrics["instantaneous_output_kbps"]
        (
            result["redis_hget_p50_seconds"],
            result["redis_hget_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_hget"])
        (
            result["redis_hgetall_p50_seconds"],
            result["redis_hgetall_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_hgetall"])
        (
            result["redis_hset_p50_seconds"],
            result["redis_hset_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_hset"])
        (
            result["redis_xadd_p50_seconds"],
            result["redis_xadd_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xadd"])
        (
            result["redis_xread_p50_seconds"],
            result["redis_xread_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xread"])
        (
            result["redis_xrevrange_p50_seconds"],
            result["redis_xrevrange_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xrevrange"])
        (
            result["redis_xtrim_p50_seconds"],
            result["redis_xtrim_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xtrim"])

        metrics = EphemeralStore.info("all")
        result["redis_ephemeral_connected_clients_total"] = metrics["connected_clients"]
        result["redis_ephemeral_used_memory_rss_total"] = metrics["used_memory_rss"]
        result["redis_ephemeral_commands_processed_total"] = metrics["total_commands_processed"]
        result["redis_ephemeral_iops"] = metrics["instantaneous_ops_per_sec"]
        result["redis_ephemeral_instantaneous_input_kbps"] = metrics["instantaneous_input_kbps"]
        result["redis_ephemeral_instantaneous_output_kbps"] = metrics["instantaneous_output_kbps"]
        (
            result["redis_ephemeral_hget_p50_seconds"],
            result["redis_ephemeral_hget_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_hget"])
        (
            result["redis_ephemeral_hgetall_p50_seconds"],
            result["redis_ephemeral_hgetall_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_hgetall"])
        (
            result["redis_ephemeral_hset_p50_seconds"],
            result["redis_ephemeral_hset_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_hset"])
        (
            result["redis_ephemeral_xadd_p50_seconds"],
            result["redis_ephemeral_xadd_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xadd"])
        (
            result["redis_ephemeral_xread_p50_seconds"],
            result["redis_ephemeral_xread_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xread"])
        (
            result["redis_ephemeral_xrevrange_p50_seconds"],
            result["redis_ephemeral_xrevrange_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xrevrange"])
        (
            result["redis_ephemeral_xtrim_p50_seconds"],
            result["redis_ephemeral_xtrim_p99_seconds"],
        ) = cls.redis_extract_p50_and_p99_seconds(metrics["latency_percentiles_usec_xtrim"])

        return result



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import glob
import pathlib
import os
from typing import Optional
from openc3.top_level import add_to_search_path
from openc3.utilities.logger import Logger
from openc3.config.config_parser import ConfigParser


class Target:
    """Target encapsulates the information about a OpenC3 target. Targets are
    accessed through interfaces and have command and telemetry definition files
    which define their access."""

    def __init__(self, target_name: str, path: os.PathLike[str], gem_path: Optional[str] = None):
        """Creates a new target by processing the target.txt file in the directory
        given by the path joined with the target_name. Records all the command
        and telemetry definition files found in the targets cmd_tlm directory.
        System uses this list and processes them using PacketConfig."""
        self.language = "python"
        self.ignored_parameters = []
        self.ignored_items = []
        self.cmd_tlm_files = []
        self.interface = None
        self.routers = []
        self.cmd_cnt = 0
        self.tlm_cnt = 0
        self.cmd_unique_id_mode = False
        self.tlm_unique_id_mode = False
        self.dir: Optional[str] = None
        self.id: Optional[str] = None
        self.filename: Optional[str] = None
        self.name = target_name.upper()

        self.get_target_dir(path, gem_path)
        self.process_target_config_file()

        # If target.txt didn't specify specific cmd/tlm files then add everything
        if len(self.cmd_tlm_files) == 0:
            self.add_all_cmd_tlm()
        else:
            self.add_cmd_tlm_partials()

    def process_file(self, filename: str):
        """Parses the target configuration file
        Args:
            filename (str) The target configuration file to parse
        """
        Logger.info(f"Processing python target definition in file '{filename}'")
        parser = ConfigParser("https://docs.openc3.com/docs/configuration/target")
        for keyword, parameters in parser.parse_file(filename):
            match keyword:
                case "LANGUAGE":
                    usage = f"{keyword} <ruby | python>"
                    parser.verify_num_parameters(1, 1, usage)
                    self.language = parameters[0].lower()

                case "REQUIRE":
                    # This keyword is deprecated in Python
                    pass

                case "IGNORE_PARAMETER" | "IGNORE_ITEM":
                    usage = f"{keyword} <{keyword.split('_')[1]} NAME>"
                    parser.verify_num_parameters(1, 1, usage)
                    if "PARAMETER" in keyword:
                        self.ignored_parameters.append(parameters[0].upper())
                    if "ITEM" in keyword:
                        self.ignored_items.append(parameters[0].upper())

                case "COMMANDS" | "TELEMETRY":
                    usage = f"{keyword} <FILENAME>"
                    parser.verify_num_parameters(1, 1, usage)
                    filename = pathlib.Path(self.dir, "cmd_tlm", parameters[0])
                    if not filename.exists():
                        raise parser.error(f"{filename} not found")
                    self.cmd_tlm_files.append(filename)

                case "CMD_UNIQUE_ID_MODE":
                    usage = keyword
                    parser.verify_num_parameters(0, 0, usage)
                    self.cmd_unique_id_mode = True

                case "TLM_UNIQUE_ID_MODE":
                    usage = keyword
                    parser.verify_num_parameters(0, 0, usage)
                    self.tlm_unique_id_mode = True

                case _:
                    # blank lines will have a None keyword and should not raise an exception
                    if keyword:
                        raise Exception(parser.error(f"Unknown keyword '{keyword}'"))

    def as_json(self):
        config = {
            "name": self.name,
            "ignored_parameters": self.ignored_parameters,
            "ignored_items": self.ignored_items,
            "cmd_tlm_files": self.cmd_tlm_files,
            "id": self.id,
        }
        if self.cmd_unique_id_mode:
            config["cmd_unique_id_mode"] = True
        if self.tlm_unique_id_mode:
            config["tlm_unique_id_mode"] = True
        return config

    def get_target_dir(self, path: os.PathLike[str], gem_path: Optional[str]):
        """Get the target directory and add the target's lib folder to the
        search path if it exists
        Args:
            path (os.PathLike[str]):
            gem_path (os.)
        """
        self.dir = gem_path if gem_path else os.path.join(path, self.name)
        lib_dir = os.path.join(self.dir, "lib")
        if os.path.exists(lib_dir):
            add_to_search_path(lib_dir, False)
        proc_dir = os.path.join(self.dir, "procedures")
        if os.path.exists(proc_dir):
            add_to_search_path(proc_dir, False)

    # Process the target's configuration file if it exists
    def process_target_config_file(self):
        self.filename = os.path.join(self.dir, "target.txt")
        if os.path.exists(self.filename):
            self.process_file(self.filename)
        else:
            self.filename = None

        id_filename = os.path.join(self.dir, "target_id.txt")
        if os.path.exists(id_filename):
            with open(id_filename) as f:
                self.id = f.read().strip()
        else:
            self.id = None

    # Automatically add all command and telemetry definitions to the list
    def add_all_cmd_tlm(self):
        cmd_tlm_files = []
        cmd_tlm_dir = os.path.join(self.dir, "cmd_tlm")
        if os.path.isdir(cmd_tlm_dir):
            # Grab All *.txt files in the cmd_tlm folder and subfolders
            for filename in glob.glob(os.path.join(cmd_tlm_dir, "**", "*.txt"), recursive=True):
                if os.path.isfile(filename):
                    cmd_tlm_files.append(filename)
            # Grab All *.xtce files in the cmd_tlm folder and subfolders
            for filename in glob.glob(os.path.join(cmd_tlm_dir, "**", "*.xtce"), recursive=True):
                if os.path.isfile(filename):
                    cmd_tlm_files.append(filename)
        cmd_tlm_files.sort()
        self.cmd_tlm_files = cmd_tlm_files

    def add_cmd_tlm_partials(self):
        """Make sure all partials are included in the cmd_tlm list for the hashing sum calculation"""
        partial_files = []
        cmd_tlm_dir = os.path.join(self.dir, "cmd_tlm")
        if os.path.isdir(cmd_tlm_dir):
            # Grab all _*.txt files in the cmd_tlm folder and subfolders
            for filename in glob.glob(os.path.join(cmd_tlm_dir, "**", "_*.txt"), recursive=True):
                partial_files.append(filename)
        partial_files.sort()
        self.cmd_tlm_files = list(dict.fromkeys(self.cmd_tlm_files + partial_files))
